
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttfile.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "genfuns.h"
#include "utils.h"
#include "lllprep.h"
#include "cmdargs.h"

static FILE *f;
const char *st_prefix;
static TTtype_list st_lexicals;
static TTtype_list st_rules;
static TTtype_list st_macros;
static TTtype_list st_allrules;
static TTterm st_spec;

#define BUFSZ 1024
extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);
static char st_buf[BUFSZ];

/* Utilities. */
static int get_arity(TTterm dis)
{
	TTterm x;
	TTtype_list xs;

	assert(TTmatchLLL_disjunction(dis, &xs));
	if (TTmatch_singleton(xs, &x) &&
	    TTmatchLLL_conjunction(x, &xs)) {
		return TTlength(xs);
	}
	return 1;
}

static void gen_fun_sym(TTterm id)
{
	LLLdump_c_fun_sym(f, id);
}

static void
gen_params_tail(const char *param_prefix, TTterm id, int arity)
{
	int i;

	for (i = 1; i <= arity; i++) {
		fprintf(f, "TTterm %s%d", param_prefix, i);
		if (i != arity) {
			fprintf(f, ", ");
		}
	}
	fprintf(f, ")");
}

static void
gen_build_fun_decl(const char *sep, TTterm id, int arity)
{
	fprintf(f, "TTterm");
	fprintf(f, "%s", sep);
	LLLdump_c_build_name(f, id);
	fprintf(f, "(");
	gen_params_tail("t", id, arity);
}

static void
gen_match_fun_decl(const char *sep, TTterm id, int arity)
{
	fprintf(f, "BOOL");
	fprintf(f, "%s", sep);
	LLLdump_c_match_name(f, id);
	fprintf(f, "(TTterm t, ");
	gen_params_tail("*t", id, arity);
}

/* Generate header (Funs.c). */
static void gen_header(void)
{

	fprintf(f, "#include <stdio.h>\n");
	fprintf(f, "#include <assert.h>\n");
	fprintf(f, "#include <sys/types.h>\n");
	fprintf(f, "#include \"ttterm.h\"\n");
	fprintf(f, "#include \"ttbuffer.h\"\n");
	fprintf(f, "#include \"ttutils.h\"\n");
	fprintf(f, "#include \"%sFuns.h\"\n", st_prefix);

	fprintf(f, "\n");
	fprintf(f, "static int init = 0;\n");
	fprintf(f, "static TTterm args[1000];\n");
	fprintf(f, "\n");
}

/* Initialization (Funs.c). */
static void gen_init_code(TTterm id, int arity)
{
	if (arity <= 0)
		return;

	fprintf(f, "        ");
	gen_fun_sym(id);
	fprintf(f, " = TTbuild_fun(\"");
	LLLdump_c_ident(f, id);
	fprintf(f, "\", %d);\n", arity);
}

static void gen_cf_init(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	gen_init_code(id, get_arity(dis));
}

static void gen_lex_init(TTterm id)
{
	gen_init_code(id, 1);
}

static void gen_init(void)
{
	fprintf(f, "void\n");
	fprintf(f, "%sinit( void )\n", st_prefix);
	fprintf(f, "{\n");
	fprintf(f, "    if(init == 0)\n");
	fprintf(f, "    {\n");
	fprintf(f, "        TTutils_init();\n");

	TTmap_void(gen_cf_init, st_rules);
	TTmap_void(gen_lex_init, st_lexicals);

	fprintf(f, "    }\n");
	fprintf(f, "    init++;\n");
	fprintf(f, "}\n\n");
}

/* Generate typechecks (Funs.c). */
static void gen_typecheck_basis(TTterm t, int n)
{
	TTterm x, u;

	assert(TTmatchLLL_basis(t, &x));
	if (TTmatchLLL_ident(x, &u)) {
		LLLdump_c_typecheck(f, x);
		fprintf(f, "(t%d)", n);
	} else {
		fprintf(stderr, "Internal error, dumping term.\n");
		TTdump(x);
		assert(0);
	}
}

static void gen_typecheck_conjunction(TTterm t)
{
	TTterm hd, tl, x;
	int i;

	assert(TTmatchLLL_conjunction(t, &tl));
	i = 1;
	while (TTmatch_cons(tl, &hd, &tl)) {
		assert(TTmatchLLL_term(hd, &hd, &x));
		assert(TTmatch_nil(x));
		gen_typecheck_basis(hd, i);
		if (!TTmatch_nil(tl)) {
			fprintf(f, " && ");
		}
		i++;
	}
}

static void gen_typecheck_dis(TTterm dis)
{
	TTtype_list tl;
	TTterm hd;
	BOOL first;

	assert(TTmatchLLL_disjunction(dis, &tl));

	first = TRUE;
	while(TTmatch_cons(tl, &hd, &tl)) {
		if (!first) {
			fprintf(f, " || ");
		} else {
			first = FALSE;
		}
		gen_typecheck_conjunction(hd);
	}
}

/* Generate is (Funs.c). */
static void gen_is_rule(TTterm t)
{
	TTterm id, der, dis, dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	if (TTequal_string(der, "=")) {
		return;
	}
	fprintf(f, "BOOL\n");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t)\n");
	fprintf(f, "{\n");
	fprintf(f, "    t = TTunsandwich(t);\n");
	fprintf(f, "    return t->fun == ");
	gen_fun_sym(id);
	fprintf(f, ";\n");
	fprintf(f, "}\n\n");
}

static void gen_lex_is_rule(TTterm id)
{
	fprintf(f, "BOOL\n");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t)\n");
	fprintf(f, "{\n");
	fprintf(f, "    t = TTunsandwich(t);\n");
	fprintf(f, "    return t->fun == ");
	gen_fun_sym(id);
	fprintf(f, ";\n");
	fprintf(f, "}\n\n");
}

static void gen_is_macro_proto(TTterm t)
{
	TTterm id, der, dis, dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	fprintf(f, "static BOOL ");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t);\n");
}

static void gen_is_macro_basis(TTterm t)
{
	TTterm x, b1, b2, u, id, rep;

	assert(TTmatchLLL_basis(t, &x));

	if (TTmatchLLL_alternation(x, &u, &b1, &b2, &u, &rep)) {
		fprintf(f, "TTtypecheck_");
		assert(TTmatchLLL_repetition(rep, &x));
		if (TTequal_string(x, "+")) {
			fprintf(f, "plus");
		} else if (TTequal_string(x, "*")) {
			fprintf(f, "star");
		} else if (TTequal_string(x, "?")) {
			fprintf(f, "optional");
		}
		fprintf(f, "(t, ");
		id = LLLget_ident(b1);
		LLLdump_c_typecheck(f, id);
		fprintf(f, ")");
	} else if (x->fun == TTfunLLL_literal) {
		fprintf(f, "TTtypecheck_literal(t, ");
		LLLdump_literal(f, x);
		fprintf(f, ")");
	}
}

static void gen_is_macro(TTterm t)
{
	TTterm id, der, dis, dot, x, xs, b1, rep, rep_opt;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	fprintf(f, "static BOOL\n");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t)\n");
	fprintf(f, "{\n");
	fprintf(f, "    t = TTunsandwich(t);\n");
	fprintf(f, "    return ");
	assert(TTmatchLLL_disjunction(dis, &xs));
	assert(TTmatch_singleton(xs, &x));
	assert(TTmatchLLL_conjunction(x, &xs));
	assert(TTmatch_singleton(xs, &x));
	if (TTmatchLLL_term(x, &b1, &rep_opt)) {
		if (TTmatch_singleton(rep_opt, &rep)) {
		    	assert(TTmatchLLL_repetition(rep, &x));
			fprintf(f, "TTtypecheck_");
			if (TTequal_string(x, "+")) {
				fprintf(f, "plus");
			} else if (TTequal_string(x, "*")) {
				fprintf(f, "star");
			} else if (TTequal_string(x, "?")) {
				fprintf(f, "optional");
			}
			fprintf(f, "(t, ");
			id = LLLget_ident(b1);
			LLLdump_c_typecheck(f, id);
			fprintf(f, ")");
		} else {
			gen_is_macro_basis(b1);
		}
	}

	fprintf(f, ";\n");
	fprintf(f, "}\n\n");
}

/* Generate builds (Funs.c). */
static void gen_cf_build(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	int i, arity;

	assert(TTmatchLLL_rule((TTterm) rule, &id, &der, &dis, &dot));

	arity = get_arity(dis);

	gen_build_fun_decl("\n", id, arity);

	fprintf(f, "\n");
	fprintf(f, "{\n");

 	fprintf(f, "    assert(");
	gen_typecheck_dis(dis);
 	fprintf(f, ");\n");

	for (i = 0; i < arity; i++) {
		fprintf(f, "    args[%d] = t%d;\n", i, i + 1);
	}
	fprintf(f, "    return ");
	fprintf(f, "TTbuild_term(");
	gen_fun_sym(id);
	fprintf(f, ", args);\n");
	fprintf(f, "}\n\n");
}

static void gen_lex_build(TTterm id)
{
	fprintf(f, "TTterm\n");
	LLLdump_c_build_name(f, id);
	fprintf(f, "(TTtype_string t)\n");
	fprintf(f, "{\n");
	fprintf(f, "    args[0] = (TTterm)t;\n");
	fprintf(f, "    return ");
	fprintf(f, "TTbuild_term(");
	gen_fun_sym(id);
	fprintf(f, ", args);\n");
	fprintf(f, "}\n\n");
}

/* Generate matches (Funs.c). */
static void gen_cf_match(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	int i, arity;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	arity = get_arity(dis);

	gen_match_fun_decl("\n", id, arity);
	fprintf(f, "\n");
	fprintf(f, "{\n");

	fprintf(f, "    if (t->fun == ");
	gen_fun_sym(id);
	fprintf(f, ")\n");
	fprintf(f, "    {\n");

	for (i = 0; i < arity; i++) {
		fprintf(f, "        *t%d = ", i + 1);
		fprintf(f, "t->u.args[%d];\n", i);
	}

	fprintf(f, "        return TRUE;\n");
	fprintf(f, "    }\n");
	fprintf(f, "    return FALSE;\n");
	fprintf(f, "}\n\n");
}

static void gen_lex_match(TTterm id)
{
	fprintf(f, "BOOL\n");
	LLLdump_c_match_name(f, id);
	fprintf(f, "(TTterm t, TTtype_string *t1)\n");
	fprintf(f, "{\n");
	fprintf(f, "    if (t->fun == ");
	gen_fun_sym(id);
	fprintf(f, ")\n");
	fprintf(f, "    {\n");
	fprintf(f, "        *t1 = t->u.args[0];\n");
	fprintf(f, "        return TRUE;\n");
	fprintf(f, "    }\n");
	fprintf(f, "    return FALSE;\n");
	fprintf(f, "}\n\n");
}

/*----------------------------------------------------------------------------*/
/* Generate Funs.c file.                                                      */
/*----------------------------------------------------------------------------*/

void generate_funs_c(void)
{
	gen_init();
	TTmap_void(gen_is_macro_proto, st_macros);
	fprintf(f, "\n");
	TTmap_void(gen_is_macro, st_macros);
	TTmap_void(gen_is_rule, st_rules);
	TTmap_void(gen_lex_is_rule, st_lexicals);
	TTmap_void(gen_cf_build, st_rules);
	TTmap_void(gen_lex_build, st_lexicals);
	TTmap_void(gen_cf_match, st_rules);
	TTmap_void(gen_lex_match, st_lexicals);
}


/*----------------------------------------------------------------------------*/
/* Function symbols (Funs.h).                                                 */
/*----------------------------------------------------------------------------*/

static void gen_static_code(TTterm id, int arity)
{
	if (arity <= 0)
		return;

	fprintf(f, "TTfun ");
	gen_fun_sym(id);
	fprintf(f, ";\n");
}

static void gen_cf_funs(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	gen_static_code(id, get_arity(dis));
}

static void gen_lex_funs(TTterm id)
{
	gen_static_code(id, 1);
}

/*----------------------------------------------------------------------------*/
/* Generate builds (Funs.h).                                                  */
/*----------------------------------------------------------------------------*/

static void gen_cf_build_h(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	gen_build_fun_decl(" ", id, get_arity(dis));
	fprintf(f, ";\n");
}

static void gen_lex_build_h(TTterm id)
{
	fprintf(f, "TTterm ");
	LLLdump_c_build_name(f, id);
	fprintf(f, "(TTtype_string t);\n");
}

/*----------------------------------------------------------------------------*/
/* Generate matches (Funs.h).                                                 */
/*----------------------------------------------------------------------------*/

static void gen_cf_match_h(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	gen_match_fun_decl(" ", id, get_arity(dis));
	fprintf(f, ";\n");
}

static void gen_lex_match_h(TTterm id)
{
	fprintf(f, "BOOL    ");
	LLLdump_c_match_name(f, id);
	fprintf(f, "(TTterm t, TTtype_string *t1);\n");
}

/*----------------------------------------------------------------------------*/
/* Generate Funs.h file.                                                      */
/*----------------------------------------------------------------------------*/
static void gen_is_rule_h(TTterm t)
{
	TTterm id, der, dis, dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	if (TTequal_string(der, "=")) {
		return;
	}
	fprintf(f, "BOOL ");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t);\n");
}

static void gen_lex_is_rule_h(TTterm id)
{
	fprintf(f, "BOOL ");
	LLLdump_c_typecheck(f, id);
	fprintf(f, "(TTterm t);\n");
}

void generate_funs_h(void)
{
	fprintf(f, "\n");
	fprintf(f, "void %sinit(void);\n", st_prefix);
	fprintf(f, "\n");

	TTmap_void(gen_cf_funs, st_rules);
	TTmap_void(gen_lex_funs, st_lexicals);
	fprintf(f, "\n");

	TTmap_void(gen_cf_build_h, st_rules);
	TTmap_void(gen_lex_build_h, st_lexicals);
	TTmap_void(gen_cf_match_h, st_rules);
	TTmap_void(gen_lex_match_h, st_lexicals);
	TTmap_void(gen_is_rule_h, st_rules);
	TTmap_void(gen_lex_is_rule_h, st_lexicals);

}

/* Generate Funs.c and Funs.h. */
void generate_funs_init(FILE * g, TTterm spec)
{
	LLLprep_comb(spec, &st_rules, &st_macros);
	st_spec = TTbuildLLL_specification(TTconcat(st_rules, st_macros));
	st_lexicals = LLLget_lexicals(spec, gv_opts.external_modules);
	st_prefix = gv_opts.prefix;

	assert(TTmatchLLL_specification(st_spec, &st_allrules));
	f = g;
}

void generate_funs(TTterm spec)
{
	generate_funs_init(f, spec);

	snprintf(st_buf, BUFSZ, "%sFuns.c", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_header();

	generate_funs_c();

	fclose(f);

	snprintf(st_buf, BUFSZ, "%sFuns.h", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	generate_funs_h();

	fclose(f);
}

