
#include <stdio.h>
#include <assert.h>
#include <sys/types.h>
#include <ctype.h>
#include <string.h>

#include "ttterm.h"
#include "ttfile.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "genhaskell.h"
#include "utils.h"
#include "cmdargs.h"

#define MAX_BUF_SIZE 1000

extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);

static FILE *f;
const char *st_prefix;
static TTtype_list st_lexicals;
static TTtype_list st_keywords;
static TTtype_list st_rules;
static TTterm st_spec;
static int st_constructor_nr;
static TTterm st_constructor_id;

static char st_buffer[MAX_BUF_SIZE];

static void gen_id(TTterm t)
{
	TTterm str;
	const char * p;
	const char * q;
	char ch;

	assert(TTmatchLLL_ident(t, &str));
	assert(TTmatch_string(str, &p));
	q = strrchr(p, '/');
	if (q == 0) {
		q = p;
	} else {
		q++;
	}
	assert(strlen(q) >= 1);
	fprintf(f, "%c", toupper(*q++));
	while ((ch = *q++) != '\0') {
		if (ch == '-') {
			fprintf(f, "_");
		} else {
			fputc(ch, f);
		}
	}
}

static void gen_literal(TTterm t)
{
	TTterm str;

	assert(TTmatchLLL_literal(t, &str));
	TTdump_string(f, str);
}

static void gen_data_basis(TTterm t)
{
	TTterm x;

	assert(TTmatchLLL_basis(t, &x));
	if (x->fun == TTfunLLL_ident) {
		gen_id(x);
	} else if (x->fun == TTfunLLL_literal) {
		fprintf(f, "Token");
	} else if (TTequal_string(t, "%epsilon")) {
		fprintf(stderr, "Not supported for Haskell generation: %%epsilon.\n");
	} else if (x->fun == TTfunLLL_alternation) {
		fprintf(stderr, "Not supported for Haskell generation: alternations.\n");
	} else if (x->fun == TTfunLLL_group) {
		fprintf(stderr, "Not supported for Haskell generation: groups.\n");
	} else {
		assert(0);
	}
}

static void gen_data_term(TTterm t)
{
	TTterm basis, rep_opt, rep, str;

	assert(TTmatchLLL_term(t, &basis, &rep_opt));
	if (TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		if (TTequal_string(str, "+") || TTequal_string(str, "*")) {
			fprintf(f, "[");
			gen_data_basis(basis);
			fprintf(f, "]");
		} else if (TTequal_string(str, "?")) {
			fprintf(f, "(Prelude.Maybe ");
			gen_data_basis(basis);
			fprintf(f, ")");
		} else {
			assert(0);
		}
	} else {
		gen_data_basis(basis);
	}
}

static void gen_data_conjunction(TTterm t)
{
	TTterm xs, x;

	assert(TTmatchLLL_conjunction(t, &xs));
	while (TTmatch_cons(xs, &x, &xs)) {
		fprintf(f, "!");
		gen_data_term(x);
		if (!TTmatch_nil(xs)) {
			fprintf(f, " ");
		}
	}
}

static void gen_data_conjunction_sep(TTterm t)
{
	fprintf(f, "\n");
	fprintf(f, "  | ");
	gen_id(st_constructor_id);
	fprintf(f, "__%d", st_constructor_nr++);
	fprintf(f, " ");
	gen_data_conjunction(t);
}

static void gen_data_rule(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	TTterm x, xs;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));
	if (TTmatch_singleton(xs, &x)) {
		fprintf(f, "data ");
		gen_id(id);
		fprintf(f, " = ");
		gen_id(id);
		fprintf(f, " ");
		gen_data_conjunction(x);
	} else {
		st_constructor_nr = 1;
		st_constructor_id = id;
		fprintf(f, "data ");
		gen_id(id);
		fprintf(f, "\n");
		fprintf(f, "  = ");
		assert(TTmatch_cons(xs, &x, &xs));
		gen_id(id);
		fprintf(f, "__%d", st_constructor_nr++);
		fprintf(f, " ");
		gen_data_conjunction(x);
		TTmap_void(gen_data_conjunction_sep, xs);
	}
	fprintf(f, "\n\n");
}

static void gen_data_lex(TTterm t)
{
	fprintf(f, "data ");
	gen_id(t);
	fprintf(f, " = ");
	gen_id(t);
	fprintf(f, " !Token");
	fprintf(f, "\n");
}

static void gen_data(void)
{
	snprintf(st_buffer, MAX_BUF_SIZE, "%sData.hs", st_prefix);
	f = TTfile_open(st_buffer, "w");
	fprintf(f, "module %sData where\n", st_prefix);
	fprintf(f, "import qualified Prelude\n");
	fprintf(f, "\n");
	fprintf(f, "data Token = Token !Prelude.String !Prelude.String !Prelude.String\n");
	fprintf(f, "\n");
	TTmap_void(gen_data_rule, st_rules);
	TTmap_void(gen_data_lex, st_lexicals);
	fclose(f);
}

static void gen_parse_basis(TTterm t)
{
	TTterm x;

	assert(TTmatchLLL_basis(t, &x));
	if (x->fun == TTfunLLL_ident) {
		fprintf(f, "parse");
		gen_id(x);
	} else if (x->fun == TTfunLLL_literal) {
		fprintf(f, "(literal ");
		gen_literal(x);
		fprintf(f, ")");
	} else if (TTequal_string(t, "%epsilon")) {
		fprintf(stderr, "Not supported for Haskell generation: %%epsilon.\n");
	} else if (x->fun == TTfunLLL_alternation) {
		fprintf(stderr, "Not supported for Haskell generation: alternations.\n");
	} else if (x->fun == TTfunLLL_group) {
		fprintf(stderr, "Not supported for Haskell generation: groups.\n");
	} else {
		assert(0);
	}
}

static void gen_parse_term(TTterm t)
{
	TTterm basis, rep_opt, rep, str;

	assert(TTmatchLLL_term(t, &basis, &rep_opt));
	if (TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		if (TTequal_string(str, "+")) {
			fprintf(f, "(many1 ");
			gen_parse_basis(basis);
			fprintf(f, ")");
		} else if (TTequal_string(str, "*")) {
			fprintf(f, "(many ");
			gen_parse_basis(basis);
			fprintf(f, ")");
		} else if (TTequal_string(str, "?")) {
			fprintf(f, "(optional ");
			gen_parse_basis(basis);
			fprintf(f, ")");
		} else {
			assert(0);
		}
	} else {
		gen_parse_basis(basis);
	}
}

static void gen_parse_conjunction(TTterm t)
{
	int n;
	TTterm x, xs;
	int i, arity;

	fprintf(f, "do");
	fprintf(f, "    {\n");
	assert(TTmatchLLL_conjunction(t, &xs));
	arity = TTlength(xs);
	n = 1;
	while (TTmatch_cons(xs, &x, &xs)) {
		fprintf(f, "        a%d <- ", n++);
		gen_parse_term(x);
		fprintf(f, ";\n");
	}
	fprintf(f, "        return (A.");
	gen_id(st_constructor_id);
	if (st_constructor_nr != 0) {
		fprintf(f, "__%d", st_constructor_nr++);
	}
	for (i = 1; i <= arity; i++) {
		fprintf(f, " a%d", i);
	}
	fprintf(f, ");\n");
	fprintf(f, "    }");
}

static void gen_parse_conjunction_sep(TTterm t)
{
	fprintf(f, "\n");
	fprintf(f, "    +++");
	fprintf(f, "\n");
	fprintf(f, "    ");
	gen_parse_conjunction(t);
}

static void gen_parse_rule(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	TTterm x, xs;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));
	fprintf(f, "parse");
	gen_id(id);
	fprintf(f, " :: Parser A.");
	gen_id(id);
	fprintf(f, "\n");
	fprintf(f, "parse");
	gen_id(id);
	fprintf(f, "\n");
	fprintf(f, "  = ");
	st_constructor_id = id;
	if (TTmatch_singleton(xs, &x)) {
		st_constructor_nr = 0;
		gen_parse_conjunction(x);
	} else {
		st_constructor_nr = 1;
		assert(TTmatch_cons(xs, &x, &xs));
		gen_parse_conjunction(x);
		TTmap_void(gen_parse_conjunction_sep, xs);
	}
	fprintf(f, "\n\n");
}

static void gen_parse(void)
{
	snprintf(st_buffer, MAX_BUF_SIZE, "%sParse.hs", st_prefix);
	f = TTfile_open(st_buffer, "w");
	fprintf(f, "module %sParse where\n", st_prefix);
	fprintf(f, "import qualified Prelude\n");
	fprintf(f, "import qualified %sData as A\n", st_prefix);
	fprintf(f, "import ParseLib\n");
	fprintf(f, "import Lex%s\n", st_prefix);
	fprintf(f, "\n");
	TTmap_void(gen_parse_rule, st_rules);
	fclose(f);
}

static void gen_lex_lex(TTterm t)
{
	fprintf(f, "parse");
	gen_id(t);
	fprintf(f, " :: Parser A.");
	gen_id(t);
	fprintf(f, "\n");
	fprintf(f, "parse");
	gen_id(t);
	fprintf(f, "\n");
	fprintf(f, "  = parse (token (do\n");
	fprintf(f, "    {\n");
	fprintf(f, "        return (A.");
	gen_id(t);
	fprintf(f, " \"TODO\");\n");
	fprintf(f, "    }))\n\n");
}

static void gen_keywords(void)
{
	TTterm xs, x, str;

	fprintf(f, "literals :: [Prelude.String]\n");
	fprintf(f, "literals = [\n");
	xs = st_keywords;
	while (TTmatch_cons(xs, &x, &xs)) {
		assert(TTmatchLLL_literal(x, &str));
		fprintf(f, "  ");
		TTdump_string(f, str);
		if (!TTmatch_nil(xs)) {
			fprintf(f, ",\n");
		}
	}
	fprintf(f, "]");
	fprintf(f, "\n\n");
}

static void gen_lex(void)
{
	snprintf(st_buffer, MAX_BUF_SIZE, "%sLex.hs", st_prefix);
	f = TTfile_open(st_buffer, "w");
	fprintf(f, "module %sLex where\n", st_prefix);
	fprintf(f, "import qualified Prelude\n");
	fprintf(f, "import qualified %sData as A\n", st_prefix);
	fprintf(f, "import ParseLib\n");
	fprintf(f, "\n");
	TTmap_void(gen_lex_lex, st_lexicals);
	gen_keywords();
	fclose(f);
}

void generate_haskell(TTterm spec)
{
	st_prefix = gv_opts.prefix;
	st_spec = spec;
	assert(TTmatchLLL_specification(st_spec, &st_rules));
	st_lexicals = LLLget_lexicals(st_spec, gv_opts.external_modules);
	st_keywords = LLLfilter_keywords(LLLget_literals(st_spec));

	gen_data();
	gen_parse();
	gen_lex();
}


