
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttfile.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "genhtml.h"

#include "cmdargs.h"

static FILE * f;
static const char *st_prefix;
static TTterm st_spec;
static TTterm st_idents;
static TTterm st_top_idents;
static TTterm st_def_idents;
static TTterm st_used_idents;
static TTterm st_undef_idents;
static TTterm st_lexicals;
static TTterm st_literals;
static TTterm st_modules;
static TTterm st_module;
static TTterm st_basis;

static void dump_ident(TTterm t)
{
	TTterm str;

	assert(TTmatchLLL_ident(t, &str));
	TTdump_string(f, str);
}

static void dump_ident_unqualified(TTterm t)
{
	TTterm str;
	const char * x;
	const char * y;

	assert(TTmatchLLL_ident(t, &str));
	assert(TTmatch_string(str, &x));
	y = strrchr(x, '/');
	assert(y != 0);
	fprintf(f, "%s", y+1);
}

static BOOL in_module(TTterm t, TTterm module)
{
	TTterm str;
	const char * x;
	const char * y;

	assert(TTmatchLLL_ident(t, &str));
	assert(TTmatch_string(str, &x));
	assert(TTmatch_string(module, &y));
	return strncmp(x, y, strlen(y)) == 0;
}

static void dump_ident_use(TTterm t)
{
	fprintf(f, "<a href=\"#");
	dump_ident(t);
	fprintf(f, "\">");

	if (in_module(t, st_module)) {
		dump_ident_unqualified(t);
	} else {
		dump_ident(t);
	}

	fprintf(f, "</a>");
}

static void dump_ident_def(TTterm t)
{

	fprintf(f, "<a name=\"");
	dump_ident(t);
	fprintf(f, "\">");
	if (in_module(t, st_module)) {
		dump_ident_unqualified(t);
	} else {
		dump_ident(t);
	}
	fprintf(f, "</a>");
}

static void dump_literal(TTterm t)
{
	TTterm str;

	assert(TTmatchLLL_literal(t, &str));
	TTdump_string(f, str);
}

static void dump_module(TTterm t)
{
	TTdump_string(f, t);
}

static void dump_module_def(TTterm t)
{
	fprintf(f, "<a name=\"");
	dump_module(t);
	fprintf(f, "\">");
	dump_module(t);
	fprintf(f, "</a>");
}

static void dump_module_use(TTterm t)
{
	fprintf(f, "<a href=\"#");
	dump_module(t);
	fprintf(f, "\">");
	dump_module(t);
	fprintf(f, "</a>");
}

static BOOL gen_all(TTterm t)
{
	TTterm rep, u, b1, b2, str;
	const char * x;

	if (TTmatchLLL_alternation(t, &u, &b1, &b2, &u, &rep)) {
		fprintf(f, "{");
		TTtraverse(b1, gen_all);
		fprintf(f, " ");
		TTtraverse(b2, gen_all);
		fprintf(f, "}");
		assert(TTmatchLLL_repetition(rep, &str));
		TTdump_string(f, str);
		return TRUE;
	} else if (TTmatchLLL_ident(t, &u)) {
		fprintf(f, " ");
		dump_ident_use(t);
		return TRUE;
	} else if (TTmatchLLL_literal(t, &u)) {
		fprintf(f, " ");
		fprintf(f, "<b>");
		dump_literal(t);
		fprintf(f, "</b>");
		return TRUE;
	} else if (TTmatch_string(t, &x)) {
		if (strcmp(x, "+") != 0 &&
		    strcmp(x, "*") != 0 &&
		    strcmp(x, "?") != 0) {
			fprintf(f, " ");
		}
		TTdump_string(f, t);
		return TRUE;
	}
	return FALSE;
}

static TTterm get_ident_use(TTterm t)
{
	if (TTequal(t, st_basis)) {
		return st_basis;
	}
	return 0;
}

static BOOL gen_used_by(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	if(TTmatchLLL_rule(t, &id, &der, &dis, &dot)) {
		if (TTselect(dis, get_ident_use)) {
			fprintf(f, " ");
			dump_ident_use(id);
		}
		return TRUE;
	}
	return FALSE;
}

static void gen_rule(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	TTterm hd, xs;

	fprintf(f, "<p>\n");

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	dump_ident_def(id);
	if (!TTelem(id, st_top_idents)) {
		fprintf(f, " <b>used by</b>");
		st_basis = id;
		TTtraverse(st_spec, gen_used_by);
	}
	fprintf(f, "<ul>\n");
	assert(TTmatchLLL_disjunction(dis, &xs));
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "<li>\n");
		TTtraverse(hd, gen_all);
	}
	fprintf(f, "</ul>\n");
}

static TTterm get_module_a(TTterm t)
{
	TTtype_string str;
	const char *x;
	const char *y;

	if (TTmatchLLL_ident(t, &str)) {
		assert(TTmatch_string(str, &x));
		y = strrchr(x, '/');
		assert(y != 0);
		return TTbuild_singleton(TTbuild_string_n(x, (size_t)(y - x)));
	}
	return 0;
}

static TTterm get_rule_a(TTterm t)
{
	TTterm id, u, str;
	const char * x;
	const char * y;

	if (TTmatchLLL_rule(t, &id, &u, &u, &u)) {
		assert(TTmatchLLL_ident(id, &str));
		assert(TTmatch_string(str, &x));
		assert(TTmatch_string(st_module, &y));
		if (strncmp(x, y, strlen(y)) == 0) {
			return TTbuild_singleton(t);
		}
		return TTbuild_nil();
	}
	return 0;
}

static TTterm get_import_a(TTterm t)
{
	TTterm str;
	const char * x;
	const char * y;

	if (TTmatchLLL_ident(t, &str)) {
		assert(TTmatch_string(str, &x));
		assert(TTmatch_string(st_module, &y));
		if (strncmp(x, y, strlen(y)) != 0) {
			y = strrchr(x, '/');
			assert(y != 0);
			return TTbuild_singleton(TTbuild_string_n(x, (size_t)(y - x)));
		}
		return TTbuild_nil();
	}
	return 0;
}

static TTterm get_export_a(TTterm t)
{
	TTterm id, u;

	if (TTmatchLLL_rule(t, &id, &u, &u, &u)) {
		return TTbuild_singleton(id);
	}
	return 0;
}

static void gen_import(TTterm t)
{
	fprintf(f, "<li>");
	dump_module_use(t);
	fprintf(f, "\n");
}

static void gen_export(TTterm t)
{
	fprintf(f, "<li>");
	dump_ident_use(t);
	fprintf(f, "\n");
}

static void gen_module(TTterm t)
{
	TTterm imports, exports, rules;

	st_module = t;
	rules = TTanalyze(st_spec, get_rule_a, TTunion, TTbuild_nil());
	imports = TTanalyze(rules, get_import_a, TTunion, TTbuild_nil());
	exports = TTanalyze(rules, get_export_a, TTunion, TTbuild_nil());

	fprintf(f, "<h2>");
	dump_module_def(t);
	fprintf(f, "</h2>\n");

	fprintf(f, "<ul>\n");
	fprintf(f, "<li> imports:\n");
	fprintf(f, "<ul>\n");
	TTmap_void(gen_import, imports);
	fprintf(f, "</ul>\n");
	fprintf(f, "<li> exports:\n");
	fprintf(f, "<ul>\n");
	TTmap_void(gen_export, exports);
	fprintf(f, "</ul>\n");
	fprintf(f, "</ul>\n");
	TTmap_void(gen_rule, rules);
}

static void gen_syntax(void)
{
	fprintf(f, "<a name=Syntax><h1>EBNF Syntax</h1></a>\n");
	TTmap_void(gen_module, st_modules);
}

static void gen_contents(void)
{
	fprintf(f, "<h1>Contents</h1>\n");
	fprintf(f, "<ul>\n");
	fprintf(f, "<li><a HREF=#Summary>Summary</a>\n");
	fprintf(f, "<li><a HREF=#Syntax>EBNF syntax</a>\n");
	fprintf(f, "<li><a HREF=#TopSorts>Top sorts</a>\n");
	fprintf(f, "<li><a HREF=#BottomSorts>Bottom sorts</a>\n");
	fprintf(f, "<li><a HREF=#Keywords>Keywords</a>\n");
	fprintf(f, "</ul>\n");
}

static void gen_summary(void)
{
	fprintf(f, "<a name=Summary><h1>Summary</h1></a>\n");
	
	fprintf(f, "<ul>\n");
	fprintf(f, "<li> Number of modules: %d\n", TTlength(st_modules));
	fprintf(f, "<li> Number of sorts: %d\n", TTlength(st_idents));
	fprintf(f, "<li> Number of top sorts: %d\n", TTlength(st_top_idents));
	fprintf(f, "<li> Number of used sorts: %d\n", TTlength(st_used_idents));
	fprintf(f, "<li> Number of defined sorts: %d\n", TTlength(st_def_idents));
	fprintf(f, "<li> Number of undefined sorts: %d\n", TTlength(st_undef_idents));
	fprintf(f, "<li> Number of lexical sorts: %d\n", TTlength(st_lexicals));
	fprintf(f, "<li> Number of literals: %d\n", TTlength(st_literals));
	fprintf(f, "</ul>\n");
}

static void gen_top_sorts(void)
{
	TTterm xs, hd;

	fprintf(f, "<a name=TopSorts><h1>Top Sorts</h1></a>\n");
	fprintf(f, "<ul>\n");
	xs = st_top_idents;
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "<li>");
		dump_ident_use(hd);
		fprintf(f, "\n");
	}
	fprintf(f, "</ul>\n");
}

static void gen_bottom_sorts(void)
{
	TTterm xs, hd;

	fprintf(f, "<a name=BottomSorts><h1>Bottom Sorts</h1></a>\n");
	fprintf(f, "<ul>\n");
	xs = st_undef_idents;
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "<li>");
		dump_ident_def(hd);
		fprintf(f, " <b>used by</b>");
		st_basis = hd;
		TTtraverse(st_spec, gen_used_by);
		fprintf(f, "\n");
	}
	fprintf(f, "</ul>\n");
}

static void gen_keywords(void)
{
	TTterm xs, hd;

	fprintf(f, "<a name=Keywords><h1>Keywords</h1></a>\n");
	fprintf(f, "<ul>\n");
	xs = st_literals;
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "<li>");
		dump_literal(hd);
		fprintf(f, " <b>used by</b>");
		st_basis = hd;
		TTtraverse(st_spec, gen_used_by);
		fprintf(f, "\n");
	}
	fprintf(f, "</ul>\n");
}

void generate_html(TTterm spec)
{
	st_spec = spec;
	st_prefix = gv_opts.prefix;

	st_modules = TTanalyze(spec, get_module_a, TTunion, TTbuild_nil());
	st_literals = LLLget_literals(spec);
	st_idents = LLLget_idents(spec);
	st_top_idents = LLLget_top_idents(spec);
	st_def_idents = LLLget_def_idents(spec);
	st_used_idents = LLLsort_idents(LLLget_used_idents(spec));
	st_undef_idents = LLLsort_idents(LLLget_undef_idents(spec));
	st_lexicals = LLLget_lexicals(spec, gv_opts.external_modules);

	f = stdout;
	gen_contents();
	gen_summary();
	gen_syntax();
	st_module = TTbuild_string("Not a module name");
	gen_top_sorts();
	gen_bottom_sorts();
	gen_keywords();
}

