
#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttfile.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "genfuns.h"
#include "genparse.h"
#include "utils.h"
#include "lllprep.h"
#include "cmdargs.h"

static FILE *f;
static TTtype_list st_top_idents;
static TTtype_list st_lexicals;
static TTtype_list st_literals;
static TTterm st_rules;
static TTterm st_macros;
static TTterm st_spec;
const char *st_prefix;

#define BUFSZ 1024
extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);
static char st_buf[BUFSZ];

/* Utilities. */
static void gen_parse_name(TTterm id)
{
	LLLdump_c_parse(f, id);
}

static BOOL is_macro(TTterm t)
{
	TTterm xs, hd;
	
	xs = st_macros;
	while (TTmatch_cons(xs, &hd, &xs)) {
		if (TTequal(t, hd)) {
			return TRUE;
		}
	}
	return FALSE;
}

/* Generate header (Parse.c). */
static void gen_header(void)
{
	fprintf(f, "#include <stdio.h>\n");
	fprintf(f, "#include <assert.h>\n");
	fprintf(f, "#include <sys/types.h>\n");
	fprintf(f, "#include \"ttterm.h\"\n");
	fprintf(f, "#include \"ttbuffer.h\"\n");
	fprintf(f, "#include \"ttutils.h\"\n");
	fprintf(f, "#include \"ttparse.h\"\n");
	fprintf(f, "#include \"%sScan.h\"\n", st_prefix);
	fprintf(f, "#include \"%sParse.h\"\n", st_prefix);
	gen_includes(f);

	fprintf(f, "\n");
	fprintf(f, "typedef TTterm(*fparse)(void);\n");
	fprintf(f, "\n");
	fprintf(f, "static int init = 0;\n");
	fprintf(f, "static TTterm args[1000];\n");
	fprintf(f, "\n");
}

/* Generate function body for macros (Parse.c). */
static BOOL gen_macro(TTterm t)
{
	TTterm b1, b2, u, rep, rep_opt, str;

	if (TTmatchLLL_alternation(t, &u, &b1, &b2, &u, &rep)) {
		fprintf(f, "    return TTparse_alternation_");
		assert(TTmatchLLL_repetition(rep, &str));
		if (TTequal_string(str, "+")) {
			fprintf(f, "plus");
		} else if (TTequal_string(str, "*")) {
			fprintf(f, "star");
		} else if (TTequal_string(str, "?")) {
			fprintf(f, "optional");
		}
		fprintf(f, "(");
		gen_parse_name(LLLget_ident(b1));
		fprintf(f, ", ");
		gen_parse_name(LLLget_ident(b2));
		fprintf(f, ");\n");
		return TRUE;
	} else if (TTmatchLLL_term(t, &b1, &rep_opt)
		   && TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		fprintf(f, "    return TTparse_");
		if (TTequal_string(str, "+")) {
			fprintf(f, "plus");
		} else if (TTequal_string(str, "*")) {
			fprintf(f, "star");
		} else if (TTequal_string(str, "?")) {
			fprintf(f, "optional");
		}
		fprintf(f, "(");
		gen_parse_name(LLLget_ident(b1));
		fprintf(f, ");\n");
		return TRUE;
	} else if (TTmatchLLL_literal(t, &u)) {
		fprintf(f, "    return TTparse%s_KEYWORD(", st_prefix);
		LLLdump_literal(f, t);
		fprintf(f, ");\n");
		return TRUE;
	}

	return FALSE;
}

/* Generate function body for conjunctions (Parse.c). */
static void gen_conjunction(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	int i, arity;
	TTterm hd, x;
	TTtype_list xs;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));
	assert(TTmatch_singleton(xs, &x));
	assert(TTmatchLLL_conjunction(x, &xs));

	arity = TTlength(xs);
	fprintf(f, "    fparse fs[%d];\n", arity);
	fprintf(f, "\n");
	i = 0;
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "    fs[%d] = ", i++);
		gen_parse_name(LLLget_ident(hd));
		fprintf(f, ";\n");
	}
	fprintf(f, "    return ");
	fprintf(f, "TTparse%s_CONJUNCTION(", st_prefix);
	LLLdump_c_fun_sym(f, id);
	fprintf(f, ", fs);\n");
}

/* Generate function body for disjunctions (Parse.c). */
static void gen_disjunction(TTterm id, TTtype_list xs)
{
	TTterm hd;
	TTtype_list ys;
	int arity, i;

	ys = xs;
	arity = TTlength(xs);

	fprintf(f, "    fparse fs[%d];\n", arity);
	i = 0;
	while (TTmatch_cons(ys, &hd, &ys)) {
		fprintf(f, "    fs[%d] = ", i++);
		gen_parse_name(LLLget_ident(hd));
		fprintf(f, ";\n");
	}
	fprintf(f, "    return ");
	fprintf(f, "TTparse%s_DISJUNCTION(", st_prefix);
	LLLdump_c_fun_sym(f, id);
	fprintf(f, ", %d, fs);\n", arity);
}

/* Generate parse functions (Parse.c). */
static BOOL gen_parse(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	TTtype_list xs, u;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)) {
		assert(TTmatchLLL_disjunction(dis, &xs));
		if (is_macro(t)) {
		fprintf(f, "static ");
		}
		fprintf(f, "TTterm\n");
		gen_parse_name(id);
		fprintf(f, "( void )\n");
		fprintf(f, "{\n");

		if (is_macro(t)) {
			TTtraverse(xs, gen_macro);
		} else if (TTmatch_singleton(xs, &u)) {
			gen_conjunction(t);
		} else {
			gen_disjunction(id, xs);
		}
		fprintf(f, "}\n\n");
		return TRUE;
	}
	return FALSE;
}

/* Generate Parse.c file. */
static BOOL gen_proto_static(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)
	    && is_macro(t)) {
		fprintf(f, "static TTterm ");
		gen_parse_name(id);
		fprintf(f, "(void);\n");
		return TRUE;
	}
	return FALSE;
}

static void gen_parse_c(void)
{
	TTtraverse(st_spec, gen_proto_static);
	fprintf(f, "\n");
	TTtraverse(st_spec, gen_parse);
}

/* Generate Parse.h file. */
static BOOL gen_proto(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)
	    && !is_macro(t)) {
		fprintf(f, "TTterm ");
		gen_parse_name(id);
		fprintf(f, "(void);\n");
		return TRUE;
	}
	return FALSE;
}

static void gen_parse_h(void)
{
	fprintf(f, "\n");
	TTtraverse(st_spec, gen_proto);
	fprintf(f, "\n");
}

/* Generate Scan.h file. */
static void gen_scan_proto(TTterm id)
{
	fprintf(f, "TTterm ");
	gen_parse_name(id);
	fprintf(f, "( void );\n");
}

static void gen_scan_h(void)
{
	fprintf(f, "\n");
	fprintf(f, "TTterm TTparse%s_LAYOUT(void);\n",
		st_prefix);
	fprintf(f, "TTterm TTparse%s_KEYWORD(const char * name);\n",
		st_prefix);
	fprintf(f, "TTterm TTparse%s_DISJUNCTION(TTfun fun, size_t n, TTterm(**parsers) (void));\n",
		st_prefix);
	fprintf(f, "TTterm TTparse%s_CONJUNCTION(TTfun fun, TTterm(**parsers) (void));\n",
		st_prefix);
	fprintf(f, "\n");
	TTmap_void(gen_scan_proto, st_lexicals);
	fprintf(f, "\n");
}

/* Generate Scan.c file. */
static void gen_scan_func(TTterm id)
{
	fprintf(f, "TTterm ");
	gen_parse_name(id);
	fprintf(f, "( void )\n");
	fprintf(f, "{\n");
	if (TTelem(id, gv_opts.rejects)) {
	}
	fprintf(f, "}\n");
	fprintf(f, "\n");
}

static void gen_scan_c(void)
{
	TTterm xs, hd;

	fprintf(f, "/* Scanner template, you can either:*/\n");
	fprintf(f, "/* - Copy this file to another file and change it.*/\n");
	fprintf(f, "/* - Change this file directly, it wont be overwritten.*/\n");
	fprintf(f, "#include <stdio.h>\n");
	fprintf(f, "#include <string.h>\n");
	fprintf(f, "#include <ctype.h>\n");
	fprintf(f, "#include <sys/types.h>\n");

	fprintf(f, "#include \"ttterm.h\"\n");
	fprintf(f, "#include \"ttbuffer.h\"\n");
	fprintf(f, "#include \"ttutils.h\"\n");
	fprintf(f, "#include \"ttparse.h\"\n");

	fprintf(f, "#include \"%sScan.h\"\n", st_prefix);
	fprintf(f, "#include \"%sParse.h\"\n", st_prefix);

	gen_includes(f);

	fprintf(f, "\n");
	fprintf(f, "/* Minimal layout definition, add comment recognition etc.*/\n");
	fprintf(f, "TTterm TTparse%s_LAYOUT(void)\n", st_prefix);
	fprintf(f, "{\n");
	fprintf(f, "    size_t len = strspn(TTparse_get_string(), \" \\n\\t\");\n");
	fprintf(f, "    return TTparse_word(len);\n");
	fprintf(f, "}\n");
	fprintf(f, "\n");
	fprintf(f, "/* There are basically two choices here:*/\n");
	fprintf(f, "/* - Keep layout in the parsetree, as is done here using \"sandwich\".*/\n");
	fprintf(f, "/* - Ignore layout, this will give a more compact parsetree, but*/\n");
	fprintf(f, "/*   preservation of layout will not be possible.*/\n");
	fprintf(f, "static TTterm scan_word(size_t len)\n");
	fprintf(f, "{\n");
	fprintf(f, "    TTterm t1, t2;\n");
	fprintf(f, "\n");
	fprintf(f, "    t1 = TTparse_word(len);\n");
	fprintf(f, "    t2 = TTparse%s_LAYOUT();\n", st_prefix);
	fprintf(f, "    return TTbuild_sandwich(TTbuild_empty_string(), t1, t2);\n");
	fprintf(f, "}\n");

	fprintf(f, "\n");
	fprintf(f, "/* Use -r to specify identifiers and this code should be ok as is. */\n");
	fprintf(f, "TTterm TTparse%s_KEYWORD(const char * name)\n", st_prefix);
	fprintf(f, "{\n");
	fprintf(f, "    int len = strlen(name);\n");
	fprintf(f, "    if (strncmp(TTparse_get_string(), name, len) == 0) {\n");
	xs = gv_opts.rejects;
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "        if (TTparse_guard(");
		gen_parse_name(TTbuildLLL_ident(hd));
		fprintf(f, ")) return 0;\n");
	}
	fprintf(f, "        return scan_word(len);\n");
	fprintf(f, "    }\n");
	fprintf(f, "    return 0;\n");
	fprintf(f, "}\n");
	fprintf(f, "\n");
	fprintf(f, "/* Passing the calls through to TTparse_disjunction and TTparse_conjunction, */\n");
	fprintf(f, "/* as is done here, should be sufficient for most languages.*/\n");
	fprintf(f, "/* See the COBOL parser for a counter example.*/\n");
	fprintf(f, "TTterm TTparse%s_DISJUNCTION(TTfun fun, size_t n, TTterm(**parsers) (void))\n", st_prefix);
	fprintf(f, "{\n");
	fprintf(f, "    return TTparse_disjunction(fun, n, parsers);\n");
	fprintf(f, "}\n");
	fprintf(f, "\n");
	fprintf(f, "TTterm TTparse%s_CONJUNCTION(TTfun fun, TTterm(**parsers) (void))\n", st_prefix);
	fprintf(f, "{\n");
	fprintf(f, "    return TTparse_conjunction(fun, parsers);\n");
	fprintf(f, "}\n");
	fprintf(f, "\n");
	if (!TTmatch_nil(st_lexicals)) {
		fprintf(f, "/* The lexical sorts have to be defined from scratch. */\n");
		TTmap_void(gen_scan_func, st_lexicals);
		fprintf(f, "\n");
	}
}

/* Generate Keyword.c file. */
/* The generated code is not very readable, but it does make a large
   difference in speed (2.63 instead of 3.46 seconds) to generate a
   O(log n) lookup instead of O(n). */
static void gen_kw(TTterm xs)
{
	int i, j, n;
	TTterm hd, ys, zs;

	n = TTlength(xs);

	assert(n != 0);

	if (n == 1) {
		fprintf(f, "        return strcmp(name, ");
		assert(TTmatch_singleton(xs, &hd));
		LLLdump_literal(f, hd);
		fprintf(f, ") == 0;\n");
		return;
	}

	i = n/2;
	j = n - i;
	ys = TTtake(i, xs);
	zs = TTdrop(i, xs);
	assert(TTlength(zs) == j);

	fprintf(f, "    if (strcmp(name, ");
	hd = TThead(zs);
	LLLdump_literal(f, hd);
	fprintf(f, ") < 0) {\n");
	gen_kw(ys);
	fprintf(f, "    } else {\n");
	gen_kw(zs);
	fprintf(f, "    }\n");
}

static void gen_keyword_c(void)
{
	fprintf(f, "\n");
	fprintf(f, "BOOL\n");
	fprintf(f, "%sisKeyword(const char * name)\n", st_prefix);
	fprintf(f, "{\n");
	if (!TTmatch_nil(st_literals)) {
		gen_kw(st_literals);
	}
	fprintf(f, "    return FALSE;\n");
	fprintf(f, "}\n");
}

/* Generate Keyword.h file. */
static void gen_keyword_h(void)
{
	fprintf(f, "\n");
	fprintf(f, "BOOL    %sisKeyword(const char * name);\n", st_prefix);
	fprintf(f, "\n");
}

/* Initialization of the combinator parser generation. */
static void init(FILE * g, TTterm spec)
{
	LLLprep_comb(spec, &st_rules, &st_macros);
	st_spec = TTbuildLLL_specification(TTconcat(st_rules, st_macros));
	st_top_idents = LLLget_top_idents(st_spec);
	st_lexicals = LLLget_lexicals(spec, gv_opts.external_modules);
	st_prefix = gv_opts.prefix;
	st_literals = LLLget_literals(st_spec);
	f = g;
}

void generate_c(TTterm spec)
{
	init(f, spec);

	snprintf(st_buf, BUFSZ, "%sParse.c", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_header();
	fprintf(f, "\n");

	gen_parse_c();
	gen_keyword_c();

	generate_funs_init(f, spec);
	generate_funs_c();

	fclose(f);

	snprintf(st_buf, BUFSZ, "%sParse.h", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_parse_h();
	gen_keyword_h();

	generate_funs_init(f, spec);
	generate_funs_h();

	fclose(f);

	snprintf(st_buf, BUFSZ, "%sScan.h", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_scan_h();

	fclose(f);

	snprintf(st_buf, BUFSZ, "%sScan.c", gv_opts.prefix);

	if (!TTfile_exists(st_buf)) {
		f = TTfile_open(st_buf, "w");
		gen_scan_c();
		fclose(f);
	}
}

