
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttfile.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "gensdf.h"
#include "lllpp.h"

#include "cmdargs.h"

#define MAX_BUF_SIZE 1000

static FILE * f;
static const char *st_prefix;
static TTterm st_spec;
static TTterm st_module;
static TTterm st_modules;
static TTterm st_idents;
static TTterm st_top_idents;
static TTterm st_def_idents;
static TTterm st_used_idents;
static TTterm st_undef_idents;
static TTterm st_lexicals;
static TTterm st_literals;

static char st_buffer[MAX_BUF_SIZE];

static void dump_ident_str(FILE *g, TTterm t)
{
	const char * p;
	const char * q;

	assert(TTmatch_string(t, &p));
	q = strrchr(p, '/');
	if (q == 0) {
		assert(p[0] != '\0');
		fputc(toupper(p[0]), g);
		fprintf(g, "%s", p+1);
	} else {
		q++;
		fputc(toupper(q[0]), g);
		fprintf(g, "%s", q+1);
	}
}

static void dump_ident(FILE *g, TTterm t)
{
	TTterm str;

	assert(TTmatchLLL_ident(t, &str));
	dump_ident_str(g, str);
}

static void build_module_name(TTterm t)
{
	const char * p;
	char * q;

	assert(TTmatch_string(t, &p));
	assert(strlen(p) < MAX_BUF_SIZE);
	strcpy(st_buffer, p);
	st_buffer[0] = toupper(st_buffer[0]);
	q = st_buffer;
	while (*q != '\0') {
		if (*q == '/') {
			*q = '-';
		}
		q++;
	}
}

static void dump_module(FILE * g, TTterm t)
{
	build_module_name(t);
	fprintf(f, "%s", st_buffer);
}

static TTterm get_module_a(TTterm t)
{
	TTtype_string str;
	const char *x;
	const char *y;

	if (TTmatchLLL_ident(t, &str)) {
		assert(TTmatch_string(str, &x));
		y = strrchr(x, '/');
		assert(y != 0);
		return TTbuild_singleton(TTbuild_string_n(x, (size_t)(y - x)));
	}
	return 0;
}

static TTterm get_rule_a(TTterm t)
{
	TTterm id, u, str;
	const char * x;
	const char * y;

	if (TTmatchLLL_rule(t, &id, &u, &u, &u)) {
		assert(TTmatchLLL_ident(id, &str));
		assert(TTmatch_string(str, &x));
		assert(TTmatch_string(st_module, &y));
		if (strncmp(x, y, strlen(y)) == 0) {
			return TTbuild_singleton(t);
		}
		return TTbuild_nil();
	}
	return 0;
}

static TTterm get_import_a(TTterm t)
{
	TTterm str;
	const char * x;
	const char * y;

	if (TTmatchLLL_ident(t, &str)) {
		assert(TTmatch_string(str, &x));
		assert(TTmatch_string(st_module, &y));
		if (strncmp(x, y, strlen(y)) != 0) {
			y = strrchr(x, '/');
			assert(y != 0);
			return TTbuild_singleton(TTbuild_string_n(x, (size_t)(y - x)));
		}
		return TTbuild_nil();
	}
	return 0;
}

static TTterm get_export_a(TTterm t)
{
	TTterm id, u;

	if (TTmatchLLL_rule(t, &id, &u, &u, &u)) {
		return TTbuild_singleton(id);
	}
	return 0;
}

static BOOL is_keyword(TTterm t)
{
	int ch;
	const char * p;

	assert(TTmatch_string(t, &p));

	while ((ch = *p++) != '\0') {
		if (isalnum(ch)) {
			return TRUE;
		}
	}
	return FALSE;
}

static void gen_import(TTterm t)
{
	fprintf(f, " ");
	dump_module(f, t);
}

static void gen_export(TTterm t)
{
	fprintf(f, "      ");
	dump_ident(f, t);
	fprintf(f, "\n");
}

static void gen_term(TTterm t);

static void gen_conjunction_nested(TTterm con)
{
	TTterm xs, hd;
	BOOL parens;

	assert(TTmatchLLL_conjunction(con, &xs));

	if (TTmatch_cons(xs, &hd, &xs)) {
		parens = !TTmatch_nil(xs);
		if (parens) {
			fprintf(f, "(");
		}
		gen_term(hd);
		while (TTmatch_cons(xs, &hd, &xs)) {
			fprintf(f, " ");
			gen_term(hd);
		}
		if (parens) {
			fprintf(f, ")");
		}
	}
}

static void gen_disjunction_nested(TTterm t)
{
	TTterm xs, hd;
	BOOL first;

	assert(TTmatchLLL_disjunction(t, &xs));

	if (TTmatch_singleton(xs, &hd)) {
		gen_conjunction_nested(hd);
	} else {
		first = TRUE;
		while (TTmatch_cons(xs, &hd, &xs)) {
			if (!first) {
				fprintf(f, " | ");
			} else {
				first = FALSE;
			}
			gen_conjunction_nested(hd);
		}
	}

}

static void gen_basis(TTterm t)
{
	TTterm x, str, u, b1, b2, dis, rep;

	assert(TTmatchLLL_basis(t, &x));

	if (TTmatchLLL_ident(x, &str)) {
		dump_ident(f, x);
	} else if (TTmatchLLL_literal(x, &str)) {
		TTdump_string(f, str);
	} else if (TTmatchLLL_alternation(x, &u, &b1, &b2, &u, &rep)) {
		fprintf(f, "{");
		gen_basis(b1);
		fprintf(f, " ");
		gen_basis(b2);
		fprintf(f, "}");
		assert(TTmatchLLL_repetition(rep, &str));
		TTdump_string(f, str);
	} else if (TTmatchLLL_group(x, &u, &dis, &u)) {
		fprintf(f, "(");
		gen_disjunction_nested(dis);
		fprintf(f, ")");
	} else {
		assert(TTequal_string(x, "%epsilon"));
		fprintf(f, "()");
	}
}

static void gen_term(TTterm t)
{
	TTterm basis, rep_opt, rep, str;

	assert(TTmatchLLL_term(t, &basis, &rep_opt));

	gen_basis(basis);
	if (TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		TTdump_string(f, str);
	}
}

static void gen_conjunction(TTterm id, TTterm con, int n)
{
	TTterm xs, term;

	assert(TTmatchLLL_conjunction(con, &xs));

	fprintf(f, "     ");
	while (TTmatch_cons(xs, &term, &xs)) {
		fprintf(f, " ");
		gen_term(term);
	}
	fprintf(f, " -> ");
	dump_ident(f, id);
	fprintf(f, " {cons(\"");
	dump_ident(f, id);
	if (n != 0) {
		fprintf(f, "-c%d", n);
	}
	fprintf(f, "\")}");
	fprintf(f, "\n");
}

static void gen_rule(TTterm t)
{
	TTterm xs, hd, id, u, dis;

	assert(TTmatchLLL_rule(t, &id, &u, &dis, &u));

	assert(TTmatchLLL_disjunction(dis, &xs));

	if (TTmatch_singleton(xs, &hd)) {
		gen_conjunction(id, hd, 0);
	} else {
		int i = 1;
		while (TTmatch_cons(xs, &hd, &xs)) {
			gen_conjunction(id, hd, i++);
		}
	}
	fprintf(f, "\n");

}

static void gen_module(TTterm t)
{
	TTterm imports, exports, rules, hd, xs;
	const char * filename;
	FILE * g;
	char ch;

	st_module = t;
	rules = TTanalyze(st_spec, get_rule_a, TTunion, TTbuild_nil());
	imports = TTanalyze(rules, get_import_a, TTunion, TTbuild_nil());
	exports = TTanalyze(rules, get_export_a, TTunion, TTbuild_nil());

	if (TTmatch_nil(exports)) {

		xs = gv_opts.includes;
		while (TTmatch_cons(xs, &hd, &xs)) {
			assert(TTmatch_string(hd, &filename));
			g = TTfile_open(filename, "r");
			while ((ch = fgetc(g)) != EOF) {
				fputc(ch, f);
			}
			fclose(g);
			fprintf(f, "\n\n");
		}

		return;
	}

	fprintf(f, "module ");
	dump_module(f, t);
	fprintf(f, "\n");

	fprintf(f, "  imports");
	TTmap_void(gen_import, imports);
	fprintf(f, "\n");
	fprintf(f, "  exports\n");
	fprintf(f, "    sorts\n");
	TTmap_void(gen_export, exports);
	fprintf(f, "\n");
	fprintf(f, "    context-free syntax\n");
	TTmap_void(gen_rule, rules);
	fprintf(f, "\n");
}

static void gen_reject(TTterm t)
{
	TTterm xs, hd;

	xs = gv_opts.rejects;

	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "      ");
		TTdump_string(f, t);
		fprintf(f, " -> ");
		dump_ident_str(f, hd);
		fprintf(f, " {reject}\n");
	}
}

static void gen_restriction(TTterm t)
{
	fprintf(f, "      ");
	TTdump_string(f, t);
	fprintf(f, "\n");
}

static TTterm lift_keyword(TTterm t)
{
	TTterm x;

	assert(TTmatchLLL_literal(t, &x));
	return x;
}

static void gen_rejects(void)
{
	TTterm keywords, xs, ys, str, id;

	keywords = TTmap(lift_keyword, st_literals);
	keywords = TTfilter(is_keyword, keywords);

	xs = gv_opts.rejects;
	ys = st_lexicals;
	while (TTmatch_cons(xs, &str, &xs)) {
		id = TTbuildLLL_ident(str);
		if (!TTelem(id, ys)) {
			fprintf(stderr, "error: cannot reject ");
			TTdump_string(stderr, str);
			fprintf(stderr, " because it is not a lexical.\n");
			exit(1);
		}
	}

	fprintf(f, "module KEYWORDS\n");
	fprintf(f, "  exports\n");
	fprintf(f, "    context-free syntax\n");

	TTmap_void(gen_reject, keywords);

	fprintf(f, "    lexical restrictions\n");

	TTmap_void(gen_restriction, keywords);

	fprintf(f, "      -/- [A-Za-z0-9\\-\\_]\n");
}

static void gen_main_import(TTterm t)
{
	fprintf(f, " ");
	dump_module(f, t);
}

static void gen_main(void)
{
	fprintf(f, "module Main\n");
	fprintf(f, "  imports");
	TTmap_void(gen_main_import, st_modules);
	fprintf(f, "\n\n");
}

static void gen_singlefile(void)
{
	f = stdout;
	fprintf(f, "%%%% Modules:     %d\n", TTlength(st_modules));
	fprintf(f, "%%%% Sorts:       %d\n", TTlength(st_idents));
	fprintf(f, "%%%% - top:       %d\n", TTlength(st_top_idents));
	fprintf(f, "%%%% - used:      %d\n", TTlength(st_used_idents));
	fprintf(f, "%%%% - defined:   %d\n", TTlength(st_def_idents));
	fprintf(f, "%%%% - undefined: %d\n", TTlength(st_undef_idents));
	fprintf(f, "%%%% - lexical:   %d\n", TTlength(st_lexicals));
	fprintf(f, "%%%% Literals:    %d\n", TTlength(st_literals));
	fprintf(f, "\n");
	fprintf(f, "definition\n");
	fprintf(f, "\n");
	gen_main();
	TTmap_void(gen_module, st_modules);
	gen_rejects();
}

static void gen_module_multifile(TTterm t)
{
	TTterm imports, exports, rules;
	static char filename[1000];

	st_module = t;
	rules = TTanalyze(st_spec, get_rule_a, TTunion, TTbuild_nil());
	imports = TTanalyze(rules, get_import_a, TTunion, TTbuild_nil());
	exports = TTanalyze(rules, get_export_a, TTunion, TTbuild_nil());

	if (TTmatch_nil(exports)) {
		return;
	}

	build_module_name(t);
	sprintf(filename, "%s.sdf", st_buffer);
	f = TTfile_open(filename, "w");

	fprintf(f, "module ");
	dump_module(f, t);
	fprintf(f, "\n");

	fprintf(f, "  imports ");
	TTmap_void(gen_import, imports);
	fprintf(f, "\n");
	fprintf(f, "  exports\n");
	fprintf(f, "    sorts\n");
	TTmap_void(gen_export, exports);
	fprintf(f, "\n");
	fprintf(f, "    context-free syntax\n");
	TTmap_void(gen_rule, rules);
	fprintf(f, "\n");

	fclose(f);
}

static void gen_rejects_multifile(void)
{
	static char filename[1000];
	sprintf(filename, "KEYWORDS.sdf");
	f = TTfile_open(filename, "w");
	gen_rejects();
	fclose(f);
}

static void gen_multifile(void)
{
	TTmap_void(gen_module_multifile, st_modules);
	gen_rejects_multifile();
}

void generate_sdf(TTterm spec)
{
	st_spec = spec;
	st_prefix = gv_opts.prefix;

	st_modules = TTanalyze(spec, get_module_a, TTunion, TTbuild_nil());
	st_literals = LLLget_literals(spec);
	st_idents = LLLget_idents(spec);
	st_top_idents = LLLget_top_idents(spec);
	st_def_idents = LLLget_def_idents(spec);
	st_used_idents = LLLsort_idents(LLLget_used_idents(spec));
	st_undef_idents = LLLsort_idents(LLLget_undef_idents(spec));
	st_lexicals = LLLget_lexicals(spec, gv_opts.external_modules);

	if (gv_opts.multifile) {
		gen_multifile();
	} else {
		gen_singlefile();
	}
}

