
#include <stdio.h>
#include <assert.h>
#include <ctype.h>
#include <string.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttfile.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "lllprep.h"
#include "utils.h"
#include "cmdargs.h"

#include "genlex.h"
#include "genfuns.h"
#include "genyacc.h"

static FILE *f;
static TTterm st_lexicals;
static TTterm st_literals;
static TTterm st_top_ident;
static TTterm st_rules;
static TTterm st_macros;
static TTterm st_spec;
const char *st_prefix;

#define BUFSZ 1024
extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);
static char st_buf[BUFSZ];

/* Generate header (Parse.y). */
static void gen_header(void)
{
	fprintf(f, "%%{\n");
	fprintf(f, "#include <ctype.h>\n");
	fprintf(f, "#include <string.h>\n");
	fprintf(f, "#include <assert.h>\n");
	fprintf(f, "#include <stdio.h>\n");
	fprintf(f, "#include <sys/types.h>\n");
	fprintf(f, "#include \"ttterm.h\"\n");
	fprintf(f, "#include \"ttbuffer.h\"\n");
	fprintf(f, "#include \"ttutils.h\"\n");
	fprintf(f, "#include \"%sFuns.h\"\n", st_prefix);
	fprintf(f, "#include \"%sParse.h\"\n", st_prefix);

	fprintf(f, "%%}\n");
}

static void gen_yacc_options(void)
{
	TTterm hd, xs;

	xs = gv_opts.yacc_options;

	while (TTmatch_cons(xs, &hd, &xs)) {
		TTdump_string(f, hd);
		fprintf(f, "\n");
	}
}

/* Generate top rule (Parse.y). */
static void gen_top(void)
{
	fprintf(f, "%%{\n");
	fprintf(f, "static TTterm st_parse_tree;\n");
	fprintf(f, "int yyparse(void);\n");
	fprintf(f, "\n");
	fprintf(f, "TTterm\n");
	LLLdump_c_parse(f, st_top_ident);
	fprintf(f, "(void)\n");
	fprintf(f, "{\n");
	fprintf(f, "    st_parse_tree = 0;\n");
	fprintf(f, "    if (yyparse() != 0) {\n");
	fprintf(f, "        return 0;\n");
	fprintf(f, "    }\n");
	fprintf(f, "    return st_parse_tree;\n");
	fprintf(f, "}\n");
	fprintf(f, "%%}\n");
	fprintf(f, "\n");
	fprintf(f, "%%start the_top_ident\n");
	gen_yacc_options();
	fprintf(f, "\n");
	fprintf(f, "%%%%\n");
	fprintf(f, "\n");
	fprintf(f, "layout\n");
	fprintf(f, "  : /* empty */ {$$ = TTbuild_empty_string();}\n");
	fprintf(f, "  | FLEX_LAYOUT {$$ = $1;}\n");
	fprintf(f, "  ;\n");
	fprintf(f, "\n");
	fprintf(f, "the_top_ident\n");
	fprintf(f, "  : layout ");
	LLLdump_c_ident(f, st_top_ident);
	fprintf(f, " { st_parse_tree = TTbuild_sandwich($1, $2, TTbuild_empty_string()); }\n");
	fprintf(f, "  ;\n");
	fprintf(f, "\n");
}

/* Generate lexical rules taking optional layout into account (Parse.y). */
static void gen_lexical_rule(TTterm t)
{
	LLLdump_c_ident(f, t);
	fprintf(f, " : ");
	LLLdump_flex_lexical(f, t);
	fprintf(f, " layout {$$ = TTbuild_sandwich(TTbuild_empty_string(), $1, $2); } ;\n\n");
}

/* Generate function body for macros (Parse.y). */
static void gen_macro(TTterm t)
{
	TTterm id, der, dis, dot, xs, x, basis, rep_opt, rep, str;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));

	assert(TTmatch_singleton(xs, &x));
	assert(TTmatchLLL_conjunction(x, &xs));
	assert(TTmatch_singleton(xs, &x));
	assert(TTmatchLLL_term(x, &basis, &rep_opt));
	if (TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		LLLdump_c_ident(f, id);
		fprintf(f, "\n");
		fprintf(f, "  : ");
		if (TTequal_string(str, "+")) {
			LLLdump_c_ident(f, LLLget_ident(basis));
			fprintf(f, " { $$ = TTbuild_singleton($1);}\n");
			fprintf(f, "  | ");
			LLLdump_c_ident(f, LLLget_ident(basis));
			fprintf(f, " ");
			LLLdump_c_ident(f, id);
			fprintf(f, " { $$ = TTbuild_cons($1, $2);}");
		} else if (TTequal_string(str, "*")) {
			fprintf(f, "/* epsilon */ { $$ = TTbuild_nil();}\n");
			fprintf(f, "  | ");
			LLLdump_c_ident(f, LLLget_ident(basis));
			fprintf(f, " ");
			LLLdump_c_ident(f, id);
			fprintf(f, " { $$ = TTbuild_cons($1, $2);}");
		} else if (TTequal_string(str, "?")) {
			fprintf(f, "/* epsilon */ { $$ = TTbuild_nil();}\n");
			fprintf(f, "  | ");
			LLLdump_c_ident(f, LLLget_ident(basis));
			fprintf(f, " { $$ = TTbuild_singleton($1);}");
		}
		fprintf(f, "\n  ;\n\n");
	}
	assert(TTmatchLLL_basis(basis, &x));
	if (x->fun == TTfunLLL_literal) {
		LLLdump_c_ident(f, id);
		fprintf(f, " : ");
		LLLdump_flex_literal(f, x);
		fprintf(f, " layout");
		fprintf(f, " { $$ = TTbuild_sandwich(TTbuild_empty_string(), $1, $2); } ;\n\n");
	}
}

/* Generate function body for conjunctions (Parse.y). */
static void gen_conjunction(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	int i, arity;
	TTterm hd, x;
	TTtype_list xs;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));
	assert(TTmatch_singleton(xs, &x));
	assert(TTmatchLLL_conjunction(x, &xs));

	arity = TTlength(xs);

	while (TTmatch_cons(xs, &hd, &xs)) {
		LLLdump_c_ident(f, LLLget_ident(hd));
		fprintf(f, "\n    ");
	}
	fprintf(f, "{ $$ = ");
	LLLdump_c_build_name(f, id);
	fprintf(f, "(");
	for (i = 1; i <= arity; i++) {
		if (i != 1) {
			fprintf(f, ", ");
		}
		fprintf(f, "$%d", i);
	}
	fprintf(f, "); }");
}

/* Generate function body for disjunctions (Parse.y). */
static void gen_disjunction(TTterm id, TTtype_list xs)
{
	TTterm hd;
	TTtype_list ys;
	BOOL first;

	ys = xs;

	first = TRUE;

	while (TTmatch_cons(ys, &hd, &ys)) {
		if (!first) {
			fprintf(f, "\n  | ");
		} else {
			first = FALSE;
		}
		LLLdump_c_ident(f, LLLget_ident(hd));
		fprintf(f, " { $$ = ");
		LLLdump_c_build_name(f, id);
		fprintf(f, "($1);}");
	}
}

/* Generate rules (Parse.y). */
static void gen_rule(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;
	TTtype_list xs, u;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	assert(TTmatchLLL_disjunction(dis, &xs));
	LLLdump_c_ident(f, id);
	fprintf(f, "\n");
	fprintf(f, "  : ");

	if (TTmatch_singleton(xs, &u)) {
		gen_conjunction(t);
	} else {
		gen_disjunction(id, xs);
	}
	fprintf(f, "\n  ;\n\n");
}

/* Generate Parse.y file. */
static void gen_y(void)
{
	gen_header();
	gen_flex_enum(f, "%token", st_lexicals);
	fprintf(f, "\n");
	gen_top();
	TTmap_void(gen_rule, st_rules);
	TTmap_void(gen_macro, st_macros);
	TTmap_void(gen_lexical_rule, st_lexicals);
}

/* Generate Parse.h file. */
static void gen_h( void )
{
	fprintf(f, "\n");
	fprintf(f, "TTterm   ");
	LLLdump_c_ident(f, st_top_ident);
	fprintf(f, "( void );\n");
	fprintf(f, "\n");
}

/* Generate Parse.y and Parse.h files. */
static void init(FILE * g, TTterm spec)
{
	LLLprep_yacc(spec, &st_rules, &st_macros);
	st_spec = TTbuildLLL_specification(TTconcat(st_rules, st_macros));
	st_lexicals = LLLget_undef_idents(st_spec);
	st_prefix = gv_opts.prefix;
	st_literals = LLLget_literals(st_spec);
	st_top_ident = LLLget_top_ident(st_spec);
	f = g;
}

void generate_yacc(TTterm spec)
{
	generate_lex(spec);
	generate_funs(spec);
	init(f, spec);

	snprintf(st_buf, BUFSZ, "%sParse.y", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_y();

	fclose(f);

	snprintf(st_buf, BUFSZ, "%sParse.h", gv_opts.prefix);
	f = TTfile_open(st_buf, "w");

	gen_h();

	fclose(f);
}

