
#include <stdio.h>
#include <string.h>
#include <assert.h>
#include <ctype.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttfile.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "SDFParse.h"

#include "cmdargs.h"
#include "genlll.h"

static TTterm st_spec;
static TTterm st_ident;
static TTterm st_module_name;
static FILE * f;

static TTterm get_ident_s(TTterm t)
{
	if (t->fun == TTfunSDF_ident) {
		return t;
	}
	return 0;
}

static TTterm get_ident(TTterm t)
{
	TTterm result = TTselect(t, get_ident_s);
	assert(result);
	return result;
}

static TTterm get_rule_ident_a(TTterm t)
{
	TTterm u, symbol;
	if (TTmatchSDF_production(t, &u, &u, &symbol, &u)) {
		return TTbuild_singleton(get_ident(symbol));
	}
	return 0;
}

static TTterm get_rule_idents(TTterm t)
{
	return TTanalyze(t, get_rule_ident_a, TTunion, TTbuild_nil());
}

static TTterm get_productions_a(TTterm t)
{
	TTterm id, u, symbol;

	if (TTmatchSDF_production(t, &u, &u, &symbol, &u)) {
		id = get_ident(symbol);
		if (TTequal(id, st_ident)) {
			return TTbuild_singleton(t);
		}
		return TTbuild_nil();
	}
	return 0;
}

static TTterm get_productions(TTterm t)
{
	st_ident = t;
	return TTanalyze(st_spec, get_productions_a, TTunion, TTbuild_nil());
}

static void gen_ident(TTterm t)
{
	TTterm str;

	assert(TTmatchSDF_ident(t, &str));
	TTdump_string(f, str);
}

static void gen_literal(TTterm t)
{
	TTterm str;

	assert(TTmatchSDF_literal(t, &str));
	TTdump_string(f, str);
}

static void gen_symbol(TTterm t);
static void gen_symbols(TTterm t);

static void gen_symbol_with_pipe(TTterm t)
{
	fprintf(f, " | ");
	gen_symbol(t);
}

static void gen_permutation_emulation(TTterm t)
{
	TTterm xs, hd;

	xs = t;
	if (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "(");
		gen_symbol(hd);
		TTmap_void(gen_symbol_with_pipe, xs);
		fprintf(f, ")*");
	} else {
		fprintf(f, "%%epsilon");
	}
}

static TTterm get_char_class(TTterm t)
{
	if (t->fun == TTfunSDF_char_class) {
		return t;
	}
	return 0;
}

static void gen_basic_symbol(TTterm t)
{
	TTterm x, u, id, lit, s1, s2, symbols;

	assert(TTmatchSDF_basic_symbol(t, &x));
	if (TTmatchSDF_sort(x, &id)) {
		gen_ident(id);
	} else if (TTmatchSDF_terminal(x, &lit)) {
		gen_literal(lit);
	} else if (TTmatchSDF_symbol_permutation(x, &u, &symbols, &u)) {
		gen_permutation_emulation(symbols);
	} else if (TTmatchSDF_symbol_group(x, &u, &symbols, &u)) {
		fprintf(f, "(");
		gen_symbols(symbols);
		fprintf(f, ")");
	} else if (TTmatchSDF_symbol_sep_plus(x, &u, &s1, &s2, &u, &u)) {
		fprintf(f, "{");
		gen_symbol(s1);
		fprintf(f, " ");
		gen_symbol(s2);
		fprintf(f, "}+");
	} else if (TTmatchSDF_symbol_sep_star(x, &u, &s1, &s2, &u, &u)) {
		fprintf(f, "{");
		gen_symbol(s1);
		fprintf(f, " ");
		gen_symbol(s2);
		fprintf(f, "}*");
	} else {
		TTdump(x);
		assert(0);
	}
}

static void gen_combined_symbol(TTterm t)
{
	TTterm basic, rep_opt, rep, str;

	assert(TTmatchSDF_combined_symbol(t, &basic, &rep_opt));
	gen_basic_symbol(basic);
	if (TTmatch_singleton(rep_opt, &rep)) {
		assert(TTmatchSDF_repetition(rep, &str));
		TTdump_string(f, str);
	}
}

static void gen_symbol(TTterm t)
{
	TTterm xs, hd;
	BOOL first;

	assert(TTmatchSDF_symbol(t, &xs));
	if (TTmatch_singleton(xs, &hd)) {
		gen_combined_symbol(hd);
	} else {
		fprintf(f, "(");
		first = TRUE;
		while(TTmatch_cons(xs, &hd, &xs)) {
			if (!first) {
				fprintf(f, " | ");
			} else {
				first = FALSE;
			}
			gen_combined_symbol(hd);
		}
		fprintf(f, ")");
	}
}

static void gen_symbol_with_space(TTterm t)
{
	fprintf(f, " ");
	gen_symbol(t);
}

static void gen_symbols(TTterm t)
{
	TTterm xs, hd;

	xs = t;
	if (TTmatch_cons(xs, &hd, &xs)) {
		gen_symbol(hd);
		TTmap_void(gen_symbol_with_space, xs);
	} else {
		fprintf(f, "%%epsilon");
	}
}

static void gen_production_body(TTterm t)
{
	TTterm symbols, u;

	assert(TTmatchSDF_production(t, &symbols, &u, &u, &u));
	gen_symbols(symbols);
}

static void gen_rule(TTterm t)
{
	TTterm xs, hd;

	xs = get_productions(t);
	/* Don't generate the rule if it contains lexical patterns. */
	if (TTselect(xs, get_char_class)) {
		return;
	}
	gen_ident(t);
	fprintf(f, "\n");
	fprintf(f, "  : ");
	assert(TTmatch_cons(xs, &hd, &xs));
	gen_production_body(hd);
	fprintf(f, "\n");
	while (TTmatch_cons(xs, &hd, &xs)) {
		fprintf(f, "  | ");
		gen_production_body(hd);
		fprintf(f, "\n");
	}
	fprintf(f, "  ;\n\n");
}

static TTterm get_module_name_s(TTterm t)
{
	TTterm u, name, str, id, symbol;

	if (TTmatchSDF_module(t, &u, &name, &u, &u)) {
		assert(TTmatchSDF_module_name(name, &id));
		assert(TTmatchSDF_ident(id, &str));
		st_module_name = TTconcat_string(str, TTbuild_string("/"));
		return 0;
	} else if (TTmatchSDF_production(t, &u, &u, &symbol, &u)) {
		id = get_ident(symbol);
		if (TTequal(id, st_ident)) {
			return st_module_name;
		}
	}
	return 0;
}

static TTterm qualify_t(TTterm t)
{
	TTterm name, id, str, qualified;

	if (TTmatchSDF_sort(t, &id)) {
		assert(TTmatchSDF_ident(id, &str));
		st_ident = id;
		name = TTselect(st_spec, get_module_name_s);
		if (name == 0) {
			name = TTbuild_string("Bottom/");
		}
		qualified = TTconcat_string(name, str);
		return TTbuildSDF_sort(TTbuildSDF_ident(qualified));
	}
	return 0;
}

void generate_lll(TTterm spec)
{
	f = stdout;
	st_spec = spec;
	st_spec = TTtransform(st_spec, qualify_t);
	TTmap_void(gen_rule, get_rule_idents(st_spec));
}

/*
SDF/sdf             : "definition" SDF/module*;
SDF/module          : "module" SDF/module-name SDF/imports-section* SDF/section*;
SDF/section         : SDF/exports-section | SDF/hiddens-section;
SDF/exports-section : "exports" SDF/grammar-section*;
SDF/hiddens-section : "hiddens" SDF/grammar-section*;
SDF/imports-section : "imports" SDF/module-name*;

SDF/grammar-section
  : SDF/lexical-syntax
    | SDF/lexical-variables
    | SDF/lexical-restrictions
    | SDF/context-free-syntax
    | SDF/context-free-variables
    | SDF/context-free-restrictions
    | SDF/context-free-priorities
    | SDF/sorts-section
   ;

SDF/lexical-syntax            : "lexical" "syntax" SDF/production*;
SDF/lexical-variables         : "lexical" "variables" SDF/production*;
SDF/lexical-restrictions      : "lexical" "restrictions" SDF/restriction*;
SDF/context-free-syntax       : "context-free" "syntax" SDF/production*;
SDF/context-free-variables    : "variables" SDF/production*;
SDF/context-free-restrictions : "context-free" "restrictions" SDF/restriction*;
SDF/context-free-priorities   : "context-free" "priorities" SDF/priorities?;
SDF/sorts-section             : "sorts" SDF/symbol*;

SDF/priorities       : {SDF/priority ","};
SDF/priority         : {SDF/production ">"};
SDF/restriction      : SDF/symbol* "-/-" SDF/char-class*;
SDF/production       : SDF/symbol* "->" SDF/symbol SDF/attributes?;
SDF/attributes       : "{" SDF/attr-seplist?  "}";
SDF/attr-seplist     : {SDF/attribute ","};
SDF/attribute        : SDF/attribute-word SDF/attribute-suffix?;
SDF/attribute-suffix : "(" SDF/literal ")";
SDF/symbol           : {SDF/combined-symbol "|"};

SDF/combined-symbol  : SDF/basic-symbol SDF/repetition?;

SDF/repetition       : "?" | "+" | "*";

SDF/basic-symbol
  : SDF/sort
  | SDF/terminal
  | SDF/char-class
  | SDF/symbol-group
  | SDF/symbol-sep-plus
  | SDF/symbol-sep-star
  | SDF/symbol-permutation
  ;

SDF/symbol-group       : "(" SDF/symbol* ")";
SDF/symbol-sep-plus    : "{" SDF/symbol SDF/symbol "}" "+";
SDF/symbol-sep-star    : "{" SDF/symbol SDF/symbol "}" "*";
SDF/symbol-permutation : "<<" SDF/symbol* ">>";


SDF/terminal : SDF/literal;
SDF/attribute-word : SDF/ident;
SDF/sort : SDF/ident;
SDF/module-name : SDF/ident;

*/


