
/* See Ralf's Grammar Adaptation paper for documentation. */
/* Also see: Framework for Syntax Transformation paper. */

#include <stdio.h>
#include <assert.h>
#include <time.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "tterror.h"

#include "LLLParse.h"
#include "FSTParse.h"
#include "lllutils.h"
#include "lllpp.h"
#include "fstpp.h"

#include "fsttrafos.h"

static TTterm st_ident;
static TTterm st_to_ident;
static TTterm st_rule;
static TTterm st_from_rule;
static TTterm st_trafo;
static TTterm st_terms;
static TTterm st_rules;

static const char * st_filename;
static int st_line;
static int st_column;

#define BUFSZ 1024

extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);
static char st_buf[BUFSZ];

/* Utilities. */

static TTterm get_ident(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	return id;
}

static TTterm get_terms(TTterm rule)
{
	TTterm u, x, xs, dis;

	assert(TTmatchLLL_rule(rule, &u, &u, &dis, &u));
	assert(TTmatchLLL_disjunction(dis, &xs));
	if (TTmatch_singleton(xs, &x)) {
		assert(TTmatchLLL_conjunction(x, &xs));
		return xs;
	} else {
		return TTbuild_singleton(TTbuildLLL_term(TTbuildLLL_basis(TTbuildLLL_group(TTbuild_string("("), dis, TTbuild_string(")"))), TTbuild_nil()));
	}
	return 0;
}

static void trafo_assert(BOOL cond, const char * msg)
{
	if (!cond) {
		snprintf(st_buf, BUFSZ, "error [%s, line %d, column %d]: %s.\n",
			st_filename, st_line, st_column, msg);
		TTerror(st_buf);
	}
}

/* The rule operators. */

static BOOL reject_ident(TTterm t)
{
	return !TTequal(get_ident(t), st_ident);
}

static TTterm delete_ident(TTterm t)
{
	TTterm x;

	if (TTmatchLLL_basis(t, &x) && TTequal(x, st_ident)) {
		return TTbuildLLL_basis(TTbuild_string("%epsilon"));
	}
	return 0;
}

static TTterm substitute_ident(TTterm t)
{
	if (TTequal(t, st_ident)) {
		return st_to_ident;
	}
	return 0;
}

static TTterm replace_ident(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis, dis2;
	TTtype_string dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	dis2 = TTtransform(dis, substitute_ident);
	if (dis2 == dis)
		return t;
	return TTbuildLLL_rule(id, der, dis2, dot);
}

static TTterm redefine_rule(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		if (TTequal(rule, st_from_rule)) {
			return st_rule;
		}
	}
	return rule;
}

static BOOL starts_with(TTterm xs, TTterm ys)
{
	TTterm x, y;

	if (TTmatch_nil(ys)) {
		return TRUE;
	}
	if (TTmatch_nil(xs)) {
		return FALSE;
	}
	assert(TTmatch_cons(xs, &x, &xs));
	assert(TTmatch_cons(ys, &y, &ys));
	if (!TTequal(x, y)) {
		return FALSE;
	}
	return starts_with(xs, ys);
}

static TTterm list_find(TTterm xs, TTterm ys)
{
	if (starts_with(xs, ys)) {
		return xs;
	}
	if (TTmatch_nil(xs)) {
		return FALSE;
	}
	return list_find(TTtail(xs), ys);
}

static TTterm fold_rule_t(TTterm t)
{
	TTterm xs, ys, pre, post, term, terms;

	if (TTmatchLLL_conjunction(t, &xs)) {
		ys = list_find(xs, st_terms);
		if (ys) {
			pre = TTtake(TTlength(xs) - TTlength(ys), xs);
			post = TTdrop(TTlength(st_terms), ys);
			term = TTbuildLLL_term(TTbuildLLL_basis(st_ident), TTbuild_nil());
			terms = TTconcat(pre, TTbuild_cons(term, post));
			return TTbuildLLL_conjunction(terms);
		}
	}
	return t;
}

static TTterm unfold_rule_t(TTterm t)
{
	TTterm x, u, dis;

	if (TTmatchLLL_basis(t, &x)) {
		if (TTequal(x, st_ident)) {
			assert(TTmatchLLL_rule(st_rule, &u, &u, &dis, &u));
			return TTbuildLLL_basis(TTbuildLLL_group(TTbuild_string("("), dis, TTbuild_string(")")));
		}
	}
	return t;
}

static TTterm add_rule(TTterm rule)
{
	TTterm id, id2;
	TTterm der, der2;
	TTterm dis, dis2, dis3;
	TTtype_string dot, dot2;
	TTtype_list xs, ys;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		assert(TTmatchLLL_disjunction(dis, &xs));
		assert(TTmatchLLL_rule
		       (st_rule, &id2, &der2, &dis2, &dot2));
		assert(TTmatchLLL_disjunction(dis2, &ys));
		dis3 =
		    TTbuildLLL_disjunction(TTunion(xs, ys));
		return TTbuildLLL_rule(id, der, dis3, dot);
	}
	return rule;
}

static TTterm sub_rule(TTterm rule)
{
	TTterm id, id2;
	TTterm der, der2;
	TTterm dis, dis2, dis3;
	TTtype_string dot, dot2;
	TTtype_list xs, ys, zs;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		assert(TTmatchLLL_disjunction(dis, &xs));
		assert(TTmatchLLL_rule
		       (st_rule, &id2, &der2, &dis2, &dot2));
		assert(TTmatchLLL_disjunction(dis2, &ys));
		zs = TTdiff(xs, ys);
		trafo_assert(!TTmatch_nil(zs), "exclude cannot be used to remove the only alternative in a production");
		dis3 = TTbuildLLL_disjunction(zs);
		return TTbuildLLL_rule(id, der, dis3, dot);
	}
	return rule;
}

/* Try to figure out if this thing can be empty, but to keep things
 * reasonably efficient (and to prevent infinite loops on cyclic
 * grammars) cutoff the search at a certain level.
 */
static BOOL possible_empty(TTterm t, int level)
{
	TTterm xs, x, y, rep_opt, rep, str, rule, u, dis;

	if (level <= 0) return FALSE;

	if (TTmatchLLL_basis(t, &x)) {
		if (TTequal_string(x, "%epsilon")) {
			return TRUE;
		}
		return possible_empty(x, level - 1);
	} else if (TTmatchLLL_term(t, &x, &rep_opt)) {
		if (TTmatch_singleton(rep_opt, &rep)) {
			assert(TTmatchLLL_repetition(rep, &str));
			if (TTequal_string(str, "+")) {
				return possible_empty(x, level - 1);
			}
			assert(TTequal_string(str, "?") ||
			       TTequal_string(str, "*"));
			return TRUE;
		} else {
			return possible_empty(x, level - 1);
		}
	} else if (t->fun == TTfunLLL_ident) {
		rule = LLLget_definition(st_rules, t);
		if (rule) {
			return possible_empty(rule, level - 1);
		}
	} else if (TTmatchLLL_rule(t, &u, &u, &dis, &u)) {
		return possible_empty(dis, level - 1);
	} else if (TTmatchLLL_disjunction(t, &xs)) {
		while (TTmatch_cons(xs, &x, &xs)) {
			if (possible_empty(x, level - 1)) {
				return TRUE;
			}
		}
	} else if (TTmatchLLL_conjunction(t, &xs)) {
		while (TTmatch_cons(xs, &x, &xs)) {
			if (!possible_empty(x, level - 1)) {
				return FALSE;
			}
		}
		return TRUE;
	} else if (TTmatchLLL_alternation(t, &u, &x, &y, &u, &rep)) {
		assert(TTmatchLLL_repetition(rep, &str));
		if (TTequal_string(str, "+")) {
			return possible_empty(x, level - 1) && possible_empty(y, level - 1);
		}
		assert(TTequal_string(str, "*"));
		return TRUE;
	} else if (TTmatchLLL_group(t, &u, &x, &u)) {
		return possible_empty(x, level - 1);
	}
	return FALSE;
}

static BOOL covers(TTterm rule1, TTterm rule2)
{
	TTterm xs, ys, u, x, y, tl;

	assert(TTmatchLLL_rule(rule1, &u, &u, &xs, &u));
	assert(TTmatchLLL_rule(rule2, &u, &u, &ys, &u));
	assert(TTmatchLLL_disjunction(xs, &xs));
	assert(TTmatchLLL_disjunction(ys, &ys));

	if (TTmatch_singleton(xs, &x) && TTmatch_singleton(ys, &y)) {
		assert(TTmatchLLL_conjunction(x, &xs));
		assert(TTmatchLLL_conjunction(y, &ys));
		while (TTmatch_cons(xs, &x, &xs)) {
			if (TTmatch_cons(ys, &y, &tl) && TTequal(x,y)) {
				ys = tl;
			} else if (!possible_empty(x, 7)) {
				return FALSE;
			}
		}
		return TTmatch_nil(ys);
	}

	return TTsubset(xs, ys);
}

static TTterm preserve_rule(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		trafo_assert(covers(st_rule, rule) && covers(rule, st_rule),
			"preserve only works on simple cases, use redefine or use more steps");
		return st_rule;
	}
	return rule;
}

static TTterm restrict_rule(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		trafo_assert(covers(rule, st_rule),
			"restrict only works on simple cases, use redefine or use more steps");
		return st_rule;
	}
	return rule;
}

static TTterm generalize_rule(TTterm rule)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	assert(TTmatchLLL_rule(rule, &id, &der, &dis, &dot));
	if (TTequal(id, st_ident)) {
		trafo_assert(covers(st_rule, rule),
			"generalize only works on simple cases, use redefine or use more steps");
		return st_rule;
	}
	return rule;
}

static TTtype_list apply_trafo(TTterm trafo, TTtype_list rules)
{
	TTterm old_rules;
	TTbuf buf;
	TTtype_string u, x;
	TTterm rule;
	TTterm ident;
	TTterm filename;
	TTterm spec;
	const char * str;

	st_trafo = trafo;

	if (TTmatchFST_eliminate(trafo, &u, &st_ident)) {
		rules = TTfilter(reject_ident, rules);
		trafo_assert(!LLLis_existing(rules, st_ident),
			"eliminate cannot be used to eliminate a rule whose identifier occurs in the right hand side of another rule");
	} else if (TTmatchFST_reject(trafo, &u, &st_ident)) {
		rules = TTfilter(reject_ident, rules);
	} else if (TTmatchFST_delete(trafo, &u, &st_ident)) {
		rules = TTtransform(rules, delete_ident);
/*		rules = TTmap(delete_ident, rules);*/
	} else if (TTmatchFST_introduce(trafo, &u, &rule)) {
		ident = get_ident(rule);
		trafo_assert(!LLLis_existing(rules, ident),
			"introduce cannot be used to introduce a rule whose identifier already exists");
		rules = TTbuild_cons(rule, rules);
	} else
	    if (TTmatchFST_rename
		(trafo, &u, &st_ident, &u, &st_to_ident)) {
		trafo_assert(LLLis_existing(rules, st_ident),
			"rename can only be used to rename an identifier that exists");
		trafo_assert(!LLLis_existing(rules, st_to_ident),
			"rename cannot be used to rename an identifier to an identifier that already exisits");
		rules = TTtransform(rules, substitute_ident);
	} else
	    if (TTmatchFST_unify
		(trafo, &u, &st_ident, &u, &st_to_ident)) {
		trafo_assert(LLLis_bottom(rules, st_ident),
			"unify can only be used to unify a bottom sort");
		trafo_assert(LLLis_existing(rules, st_to_ident),
			"unify can only be used to unify to a sort that already exists");
		rules = TTmap(replace_ident, rules);
	} else
	    if (TTmatchFST_separate
		(trafo, &u, &st_ident, &u, &st_to_ident)) {
		trafo_assert(!LLLis_existing(rules, st_to_ident),
			"separate cannot be used to separate to an identifier that already exists");
		rules = TTmap(replace_ident, rules);
		/* Maybe this assert should reflect the message, i.e. make
		   it a precondition instead of a postcondition ? */
		trafo_assert(LLLis_existing(rules, st_ident),
			"separate cannot be used on bottom sorts");
	} else if (TTmatchFST_include(trafo, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(LLLis_defined(rules, st_ident),
			"include can only be used on an existing rule");
		rules = TTmap(add_rule, rules);
	} else if (TTmatchFST_exclude(trafo, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(LLLis_defined(rules, st_ident),
			"exclude can only be used on an existing rule");
		rules = TTmap(sub_rule, rules);
	} else if (TTmatchFST_resolve(trafo, &u, &rule)) {
		ident = get_ident(rule);
		trafo_assert(LLLis_bottom(rules, ident),
			"resolve can only be used to resolve a bottom identifier");
		rules = TTbuild_cons(rule, rules);
	} else if (TTmatchFST_preserve(trafo, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(LLLis_defined(rules, st_ident),
			"preserve can only be used on an existing rule");
		rules = TTmap(preserve_rule, rules);
	} else if (TTmatchFST_restrict(trafo, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(LLLis_defined(rules, st_ident),
			"restrict can only be used on an existing rule");
		rules = TTmap(restrict_rule, rules);
	} else if (TTmatchFST_generalize(trafo, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(LLLis_defined(rules, st_ident),
			"generalize can only be used on an existing rule");
		rules = TTmap(generalize_rule, rules);
	} else if (TTmatchFST_redefine(trafo, &u, &st_from_rule, &u, &st_rule)) {
		st_ident = get_ident(st_rule);
		trafo_assert(TTequal (st_ident, get_ident(st_from_rule)),
			"redefine can only be used to redefine a rule with some identifier to a rule with the same identifier");
		trafo_assert(LLLis_defined(rules, st_ident),
			"redefine can only be used on an existing rule");
		rules = TTmap(redefine_rule, rules);
	} else if (TTmatchFST_fold(trafo, &u, &rule)) {
		st_ident = get_ident(rule);
		st_terms = get_terms(rule);
		trafo_assert(st_terms != 0,
			"fold can only be used to fold simple constructs at the moment");
		old_rules = rules;
		rules = TTtransform_nonstop(rules, fold_rule_t);
		trafo_assert(old_rules != rules,
			"fold could not applied anywhere in the grammar");
		if (!LLLis_defined(rules, st_ident)) {
			rules = TTbuild_cons(rule, rules);
		}
	} else if (TTmatchFST_unfold(trafo, &u, &st_ident)) {
		st_rule = LLLget_definition(rules, st_ident);
		trafo_assert(st_rule != 0,
			"unfold can only be used on an existing rule");
		if (st_rule) {
			rules = TTtransform_nonstop(rules, unfold_rule_t);
		}
	} else if (TTmatchFST_write(trafo, &u, &filename)) {
		assert(TTmatchLLL_literal(filename, &x));
		x = TTunquote(x);
		assert(TTmatch_string(x, &str));
		buf = TTbuffer_create();
		spec = TTbuildLLL_specification(rules);
		LLLpretty_with_header(buf, spec);
		TTbuffer_write_named(buf, str);
		TTbuffer_destroy(buf);
	} else {
		assert(0);
	}
	return rules;
}

/*
static TTtype_list apply_trafos_to_rules(TTtype_list trafos, TTtype_list rules)
{
	TTbuf buf;
	TTtype_list old_rules;
	TTterm trafo;

	while (TTmatch_cons(trafos, &trafo, &trafos)) {
		old_rules = rules;
		assert(TTmatchFST_trafo(trafo, &trafo));
		rules = apply_trafo(trafo, rules);
		if (TTequal(rules, old_rules)) {
			buf = TTbuffer_create();
			TTbuffer_append(buf, "error: unapplied transformation, dumping source:\n");
			FSTpretty(buf, trafo);
			TTbuffer_append(buf, "\n");
			TTerror(buf->data);
			TTbuffer_destroy(buf);
		}
	}
	assert(TTmatch_nil(trafos));
	return rules;
}
*/

static void apply_trafo_t(TTterm t, int line, int column)
{
	TTterm x, old_rules;

	if (TTmatchFST_trafo(t, &x)) {
		x = TTunsandwich_rec(x);
		st_line = line;
		st_column = column;
		old_rules = st_rules;
		st_rules = apply_trafo(x, st_rules);
		if (TTequal(st_rules, old_rules)) {
			snprintf(st_buf, BUFSZ, "error [%s, line %d, column %d]: unapplied transformation.\n",
				st_filename, st_line, st_column);
			TTerror(st_buf);
		}
	}
}

TTterm
FSTapply_trafos(const char * filename, TTtype_list trafos, TTterm grammar)
{
	st_filename = filename;

	assert(TTmatchLLL_specification(grammar, &st_rules));

	TTtraverse_line(trafos, apply_trafo_t, 1, 1);

	st_rules = LLLsimplify(st_rules);

	return TTbuildLLL_specification(st_rules);
}

