
#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "LLLParse.h"
#include "lllutils.h"
#include "lllprep.h"

#define BUFSZ 1024

static char st_buf[BUFSZ];

static TTtype_list st_macros;
static TTtype_list st_new_rules;
static int st_rule_id;

extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);

/* Utilities. */
static TTtype_string get_str(TTterm t)
{
	TTterm x;

	assert(TTmatchLLL_basis(t, &x));
	assert(TTmatchLLL_ident(x, &x));
	return x;
}

static void addmacro(TTterm id, TTterm t)
{
	TTterm x, rule, is, dis, dot;

	x = TTbuildLLL_conjunction(TTbuild_singleton(t));
	dis = TTbuildLLL_disjunction(TTbuild_singleton(x));
	is = TTbuild_string(":");
	dot = TTbuild_string(";");
	rule = TTbuildLLL_rule(id, is, dis, dot);
	if (!TTelem(rule, st_macros)) {
		st_macros = TTbuild_cons(rule, st_macros);
	}
}

static TTterm get_suffix(TTterm t)
{
	TTterm str;

	assert(TTmatchLLL_repetition(t, &str));
	if (TTequal_string(str, "+")) {
		return TTbuild_string("-plus");
	} else if (TTequal_string(str, "*")) {
		return TTbuild_string("-star");
	} else {
		assert(TTequal_string(str, "?"));
		return TTbuild_string("-opt");
	}
}

/* Convert to flat LLL. */
static TTterm fold_group_t(TTterm t)
{
	TTterm u, dis, id, rule;

	if (TTmatchLLL_group(t, &u, &dis, &u)) {
		snprintf(st_buf, BUFSZ, "intern/phrase%d", st_rule_id++);
		id = TTbuildLLL_ident(TTbuild_string(st_buf));
		rule = TTbuildLLL_rule(id, TTbuild_string(":"),
			 dis, TTbuild_string(";"));
		st_new_rules = TTbuild_cons(rule, st_new_rules);
		return id;
	}
	return 0;
}

static TTterm remove_groups(TTterm rules)
{
	do {
		st_new_rules = TTbuild_nil();
		rules = TTtransform(rules, fold_group_t);
		rules = TTconcat(st_new_rules, rules);
	} while (!TTmatch_nil(st_new_rules));
	return rules;
}

static TTterm remove_from_con_t(TTterm t)
{
	TTterm xs, x, id, dis, rule;

	if (TTmatchLLL_conjunction(t, &xs)) {
		if (TTmatch_singleton(xs, &x)) {
			return t;
		}
		snprintf(st_buf, BUFSZ, "intern/phrase%d", st_rule_id++);
		id = TTbuildLLL_ident(TTbuild_string(st_buf));
		dis = TTbuildLLL_disjunction(TTbuild_singleton(t));
		rule = TTbuildLLL_rule(id, TTbuild_string(":"),
			 dis, TTbuild_string(";"));
		st_new_rules = TTbuild_cons(rule, st_new_rules);
		return TTbuildLLL_conjunction(TTbuild_singleton(TTbuildLLL_term(TTbuildLLL_basis(id), TTbuild_nil())));
	}
	return 0;
}

static TTterm remove_from_dis_t(TTterm t)
{
	TTterm xs, u;

	if (TTmatchLLL_disjunction(t, &xs)) {
		if (TTmatch_singleton(xs, &u)) {
			return t;
		}
		return TTtransform(t, remove_from_con_t);
	}
	return 0;
}

static TTterm remove_anonymous_conjunctions(TTterm rules)
{
	st_new_rules = TTbuild_nil();
	rules = TTtransform(rules, remove_from_dis_t);
	rules = TTconcat(st_new_rules, rules);
	return rules;
}

static TTterm remove_epsilon_t(TTterm t)
{
	TTterm x, id;
	const char * str;

	if (TTmatchLLL_basis(t, &x)) {
		if (!TTmatch_string(x, &str)) {
			return t;
		}
		assert(strcmp(str, "%epsilon") == 0);
		id = TTbuildLLL_ident(TTbuild_string("intern/epsilon"));
		return TTbuildLLL_basis(id);
	}
	return 0;
}

static TTterm remove_epsilons(TTterm rules)
{
	return TTtransform(rules, remove_epsilon_t);
}

/* Flatten complex nested constructs. */
TTterm LLLprep_flatten(TTterm t)
{
	TTterm rules;

	st_rule_id = 1;

	assert(TTmatchLLL_specification(t, &rules));
	rules = remove_groups(rules);
	rules = remove_anonymous_conjunctions(rules);
	rules = remove_epsilons(rules);
	return TTbuildLLL_specification(rules);
}

/* Preparation for combinator parsing. */
static TTterm comb_t(TTterm t)
{
	TTterm term, basis, rep, rep_opt;
	TTtype_string str;
	TTterm id;
	TTterm b1, b2;
	TTterm l, r;

	if (TTmatchLLL_term(t, &basis, &rep_opt)
	    && TTmatch_singleton(rep_opt, &rep)) {
		basis = TTtransform(basis, comb_t);
		str = TTconcat_string(get_str(basis), get_suffix(rep));
		id = TTbuildLLL_ident(str);
		addmacro(id, TTbuildLLL_term(basis, rep_opt));
		return TTbuildLLL_term(TTbuildLLL_basis(id), TTbuild_nil());
	} else if (TTmatchLLL_alternation(t, &l, &b1, &b2, &r, &rep)) {
		b1 = TTtransform(b1, comb_t);
		b2 = TTtransform(b2, comb_t);
		str = TTconcat_string(get_str(b1),
			       TTconcat_string(TTbuild_string("-sepby-"),
					get_str(b2)));
		id = TTbuildLLL_ident(str);

		term = TTbuildLLL_alternation(l, b1, b2, r, rep);
		term = TTbuildLLL_basis(term);
		term = TTbuildLLL_term(term, TTbuild_nil());

		addmacro(id, term);

		return id;
	} else if (t->fun == TTfunLLL_literal) {
		id = LLLliteral_to_ident(t);
		term = TTbuildLLL_basis(t);
		term = TTbuildLLL_term(term, TTbuild_nil());
		addmacro(id, term);
		return id;
	}
	return 0;
}

void LLLprep_comb(TTterm t, TTterm * rules, TTterm * macros)
{
	TTterm result;

	st_macros = TTbuild_nil();
	result = t;
	result = TTtransform(result, comb_t);
	assert(TTmatchLLL_specification(result, rules));
	*macros = st_macros;
}

/* Preparation for YACC parsing. */
static TTterm yacc_t(TTterm t)
{
	TTterm term, basis, rep, rep_opt;
	TTtype_string str;
	TTterm id;
	TTterm b1, b2;
	TTterm l, r;

	if (TTmatchLLL_term(t, &basis, &rep_opt)
	    && TTmatch_singleton(rep_opt, &rep)) {
		basis = TTtransform(basis, yacc_t);
		str = TTconcat_string(get_str(basis), get_suffix(rep));
		id = TTbuildLLL_ident(str);
		addmacro(id, TTbuildLLL_term(basis, rep_opt));
		return TTbuildLLL_term(TTbuildLLL_basis(id), TTbuild_nil());
	} else if (TTmatchLLL_alternation(t, &l, &b1, &b2, &r, &rep)) {
		b1 = TTtransform(b1, yacc_t);
		b2 = TTtransform(b2, yacc_t);
		str = TTconcat_string(get_str(b1),
			       TTconcat_string(TTbuild_string("-sepby-"),
					get_str(b2)));
		id = TTbuildLLL_ident(str);

		term = TTbuildLLL_alternation(l, b1, b2, r, rep);
		term = TTbuildLLL_basis(term);
		term = TTbuildLLL_term(term, TTbuild_nil());

		addmacro(id, term);

		return id;
	} else if (TTmatchLLL_literal(t, &str)) {
		id = LLLliteral_to_ident_upper(t);
		if (TTstrlen(str) == 3) {
			term = TTbuildLLL_basis(t);
			term = TTbuildLLL_term(term, TTbuild_nil());
			addmacro(id, term);
		}
		return id;
	}
	return 0;
}

void LLLprep_yacc(TTterm t, TTterm * rules, TTterm * macros)
{
	TTterm result;

	st_macros = TTbuild_nil();
	result = t;
	result = TTtransform(result, yacc_t);
	assert(TTmatchLLL_specification(result, rules));
	*macros = st_macros;
}

