#include <stdio.h>
#include <ctype.h>
#include <assert.h>
#include <string.h>
#include <sys/types.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"

#include "LLLParse.h"
#include "lllutils.h"

#define MAX_STR_SIZE 10000
static char st_buf[MAX_STR_SIZE + 1];

static TTterm st_existing_id;
static TTterm st_defined_id;
static TTterm st_id;
static TTterm st_rule;
static const char * st_filename;
static TTterm st_external_modules;
static int st_nr_errors;

static const char * st_module_name;

extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);

static TTtype_list get_idents_a(TTterm t)
{
	if (t->fun == TTfunLLL_ident) {
		return TTbuild_singleton(t);
	}
	return (TTtype_list) 0;
}

TTtype_list LLLget_idents(TTterm t)
{
	return TTanalyze(t, get_idents_a, TTunion, TTbuild_nil());
}

static TTtype_list get_def_idents_a(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)) {
		return TTbuild_singleton(id);
	}
	return (TTtype_list) 0;
}

TTtype_list LLLget_def_idents(TTterm t)
{
	return TTanalyze(t, get_def_idents_a, TTunion, TTbuild_nil());
}

static TTtype_list get_used_idents_a(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)) {
		return TTanalyze(dis, get_used_idents_a,
			       TTunion, TTbuild_nil());
	} else if (t->fun == TTfunLLL_ident) {
		return TTbuild_singleton(t);
	}
	return (TTtype_list) 0;
}

TTtype_list LLLget_used_idents(TTterm t)
{
	return TTanalyze(t, get_used_idents_a, TTunion, TTbuild_nil());
}

TTtype_list LLLget_undef_idents(TTterm t)
{
	TTtype_list xs, ys;
	TTtype_list diff;

	xs = LLLget_idents(t);
	ys = LLLget_def_idents(t);
	diff = TTdiff(xs, ys);
	return diff;
}

static BOOL is_lexical(TTterm t)
{
	TTterm xs, hd;
	TTtype_string str;
	const char * p;
	const char * q;

	assert(TTmatchLLL_ident(t, &str));
	assert(TTmatch_string(str, &p));

	/* Check if the identifier is defined in an external module. */
	xs = st_external_modules;
	while (TTmatch_cons(xs, &hd, &xs)) {
		assert(TTmatch_string(hd, &q));
		if (strncmp(p, q, strlen(q)) == 0) {
			return FALSE;
		}
	}
	return TRUE;
}

/* A lexical is a bottom identifier that is not part of an external module. */
TTtype_list LLLget_lexicals(TTterm t, TTterm external_modules)
{
	st_external_modules = external_modules;
	return TTfilter(is_lexical, LLLget_undef_idents(t));
}

static BOOL is_qualified(TTterm t)
{
	TTtype_string str;
	const char * p;

	assert(TTmatchLLL_ident(t, &str));
	assert(TTmatch_string(str, &p));
	return strchr(p, '/') != 0;
}

TTtype_list LLLget_external(TTterm t)
{
	return TTfilter(is_qualified, LLLget_undef_idents(t));
}

static TTtype_list get_literals_a(TTterm t)
{
	TTtype_string u;

	if (TTmatchLLL_literal(t, &u)) {
		return TTbuild_singleton(t);
	}
	return (TTtype_list) 0;
}

static TTterm insertkw(TTterm x, TTterm ys)
{
	TTterm y, ys2, strx, stry;
	const char * px;
	const char * py;

	if (TTmatch_nil(ys)) {
		return TTbuild_singleton(x);
	}
	assert(TTmatch_cons(ys, &y, &ys2));
	assert(TTmatchLLL_literal(x, &strx));
	assert(TTmatchLLL_literal(y, &stry));
	assert(TTmatch_string(strx, &px));
	assert(TTmatch_string(stry, &py));
	if (strcmp(px, py) < 0) {
		return TTbuild_cons(x, ys);
	}
	return TTbuild_cons(y, insertkw(x, ys2));
}

static TTterm sortkw(TTterm t)
{
	TTterm xs, ys, x;

	xs = t;
	ys = TTbuild_nil();

	while (TTmatch_cons(xs, &x, &xs)) {
		ys = insertkw(x, ys);
	}
	return ys;
}

TTtype_list LLLget_literals(TTterm t)
{
	TTterm xs;

	xs = TTanalyze(t, get_literals_a, TTunion, TTbuild_nil());
	return sortkw(xs);
}

static BOOL is_keyword(TTterm t)
{
	TTterm x;
	int ch;
	const char * p;

	assert(TTmatchLLL_literal(t, &x));
	assert(TTmatch_string(x, &p));

	while ((ch = *p++) != '\0') {
		if (isalnum(ch)) {
			return TRUE;
		}
	}
	return FALSE;
}

TTtype_list LLLfilter_keywords(TTtype_list t)
{
	return TTfilter(is_keyword, t);
}

TTtype_list LLLget_top_idents(TTterm t)
{
	TTtype_list xs, ys, topids;

	xs = LLLget_def_idents(t);
	ys = LLLget_used_idents(t);
	topids = TTdiff(xs, ys);
	return topids;
}

/* Some generators only allow one topid (YACC for example). */
TTterm LLLget_top_ident(TTterm t)
{
	TTterm hd;
	TTtype_list tl;

	assert(TTmatch_cons(LLLget_top_idents(t), &hd, &tl));
	return hd;
}

/* all ids. */
static TTterm is_existing_select(TTterm t)
{
	if (TTequal(t, st_existing_id)) {
		return t;
	}
	return 0;
}

BOOL LLLis_existing(TTtype_list rules, TTterm id)
{
	st_existing_id = id;
	return TTselect(rules, is_existing_select) != 0;
}

/* defined ids. */
static TTterm is_defined_select(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot) &&
	    TTequal(id, st_defined_id)) {
		return t;
	}
	return 0;
}

BOOL LLLis_defined(TTtype_list rules, TTterm id)
{
	st_defined_id = id;
	return TTselect(rules, is_defined_select) != 0;
}

/* bottom ids. */
BOOL LLLis_bottom(TTtype_list rules, TTterm id)
{
	return (!LLLis_defined(rules, id) && LLLis_existing(rules, id));
}

BOOL LLLis_macro(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTterm dot;

	assert(TTmatchLLL_rule(t, &id, &der, &dis, &dot));
	return (TTequal_string(der, "="));
}

/* rules. */
static BOOL getRuleDefinitionT(TTterm t)
{
	TTterm id;
	TTterm der;
	TTterm dis;
	TTtype_string dot;

	if (TTmatchLLL_rule(t, &id, &der, &dis, &dot)) {
		if (TTequal(id, st_id)) {
			st_rule = t;
		}
		return TRUE;
	}
	return FALSE;
}

TTterm
LLLget_definition(TTterm spec, TTterm id)
{
	st_id = id;
	st_rule = 0;

	TTtraverse(spec, getRuleDefinitionT);
	return st_rule;
}

static void no_group(TTterm t, int line, int col)
{
	if (t->fun == TTfunLLL_group) {
		fprintf(stderr, "error [%s, line %d, col %d]: found group operator.\n", st_filename, line, col);
		st_nr_errors++;
	}
}

/* consistency checks. */
void LLLcheck(const char * filename, TTterm t)
{
	assert(t);
	st_filename = filename;
	st_nr_errors = 0;
	TTtraverse_line(t, no_group, 1, 1);
	if (st_nr_errors > 0) {
		fprintf(stderr, "Found %d errors, aborting.\n", st_nr_errors);
		exit(1);
	}
}

static TTterm force_qualification(TTterm t)
{
	TTterm str;
	const char * p;

	if (TTmatchLLL_ident(t, &str)) {
		assert(TTmatch_string(str, &p));
		if (strchr(p, '/') == 0) {
			snprintf(st_buf, MAX_STR_SIZE, "%s/%s", st_module_name, p);
			return TTbuildLLL_ident(TTbuild_string(st_buf));
		}
		return t;
	}
	return 0;
}

TTterm LLLforce_qualification(TTterm t, const char * module_name)
{
	st_module_name = module_name;
	return TTtransform(t, force_qualification);
}

/* Convert literals to idents. */
static const char * charname(int ch)
{
	switch (ch) {
	case ':':
		return "qcolon";
	case ';':
		return "qsemicolon";
	case '|':
		return "qpipe";
	case '_':
		return "qunderscore";
	case '?':
		return "qquestion";
	case '{':
		return "qcbopen";
	case '}':
		return "qcbclose";
	case '\'':
		return "qsingle-quote";
	case '"':
		return "qdouble-quote";
	case '\\':
		return "qbackslash";
	case '.':
		return "qdot";
	case ',':
		return "qcomma";
	case '(':
		return "qpopen";
	case ')':
		return "qpclose";
	case '<':
		return "qlt";
	case '>':
		return "qgt";
	case '=':
		return "qequal";
	case '-':
		return "qminus";
	case '+':
		return "qplus";
	case '*':
		return "qmultiply";
	case '/':
		return "qdivide";
	case '!':
		return "qexclamation";
	case '%':
		return "qpercent";
	case '[':
		return "qsbopen";
	case ']':
		return "qsbclose";
	case '@':
		return "qat";
	case '#':
		return "qhash";
	case '$':
		return "qdollar";
	case '^':
		return "qhat";
	case '&':
		return "qampersand";
	case ' ':
		return "qspace";
	default:
		fprintf(stderr, "No matching for `%c' (%d).\n", ch, ch);
		assert(0);
	}
	assert(0);
}

static void literal_to_ident(const char * name)
{
	int ch, i, j;
	const char *p;
	char *q;
	const char * tmp;
	BOOL first;
	size_t len;

	q = st_buf;
	p = name;

	i = 0;
	first = TRUE;
	q[i++] = 't';
	q[i++] = '-';
	for (j = 1; j < strlen(p) - 1; j++) {
		ch = p[j];
		if (isalnum(ch) || ch == '-' || ch == '_') {
			assert(i < MAX_STR_SIZE);
			q[i++] = ch;
		} else {
			tmp = charname(ch);
			len = strlen(tmp);
			if (!first) {
				assert(i < MAX_STR_SIZE);
				q[i++] = '-';
			}
			assert(i + len <= MAX_STR_SIZE);
			strcpy(q+i, tmp);
			i += len;
		}
		first = FALSE;
	}
	q[i] = '\0';
}

TTterm LLLliteral_to_ident(TTterm t)
{
	TTterm str;
	const char * p;

	assert(TTmatchLLL_literal(t, &str));
	assert(TTmatch_string(str, &p));
	literal_to_ident(p);
	return TTbuildLLL_ident(TTbuild_string(st_buf));
}

TTterm LLLliteral_to_ident_upper(TTterm t)
{
	TTterm str;
	const char * p;
	char * q;

	assert(TTmatchLLL_literal(t, &str));
	assert(TTmatch_string(str, &p));
	literal_to_ident(p);
	q = st_buf;
	while (*q != '\0') {
		*q = toupper(*q);
		q++;
	}
	return TTbuildLLL_ident(TTbuild_string(st_buf));
}

TTterm LLLget_ident(TTterm t)
{
	TTterm x, y;
	TTtype_list xs;


	if (t->fun == TTfunLLL_ident) {
		return t;
	} else if (TTmatchLLL_disjunction(t, &xs) ||
		   TTmatchLLL_conjunction(t, &xs)) {
		return LLLget_ident(xs);
	} else if (TTmatch_singleton((TTtype_list) t, &x)) {
		return LLLget_ident(x);
	} else if (TTmatchLLL_basis(t, &x)) {
		return LLLget_ident(x);
	} else if (TTmatchLLL_term(t, &x, &y) && TTmatch_nil(y)) {
		return LLLget_ident(x);
	}
	TTdump(t);
	assert(0);
}

static const char * get_ident_str(TTterm id)
{
	TTtype_string str;
	const char * p;

	assert(TTmatchLLL_ident(id, &str));
	assert(TTmatch_string(str, &p));
	return p;
}

static const char * get_literal_str(TTterm t)
{
	TTtype_string str;
	const char * p;

	assert(TTmatchLLL_literal(t, &str));
	assert(TTmatch_string(str, &p));
	return p;
}

static void dump_c_ident_upper(FILE * f, TTterm id)
{
	const char * p;
	int ch;

	p = get_ident_str(id);

	while ((ch = *p++) != '\0') {
		if (isalnum(ch)) {
			fputc(toupper(ch), f);
		} else {
			assert(ch == '_' || ch == '-' || ch == '/');
			fputc('_', f);
		}
	}
}

void LLLdump_c_ident(FILE * f, TTterm id)
{
	const char * p;
	int ch;

	p = get_ident_str(id);

	while ((ch = *p++) != '\0') {
		if (isalnum(ch)) {
			fputc(ch, f);
		} else {
			assert(ch == '_' || ch == '-' || ch == '/');
			fputc('_', f);
		}
	}
}

void LLLdump_flex_literal(FILE * f, TTterm t)
{
	TTterm id;
	const char * p;

	p = get_literal_str(t);

	if (strlen(p) == 3) {
		fprintf(f, "'%c'", p[1]);
	} else {
		id = LLLliteral_to_ident(t);
		fprintf(f, "FLEX_");
		dump_c_ident_upper(f, id);
	}
}

void LLLdump_flex_lexical(FILE * f, TTterm id)
{
	fprintf(f, "FLEX_");
	dump_c_ident_upper(f,id);
}

void LLLdump_literal(FILE * f, TTterm t)
{
	int ch;
	const char * p;

	p = get_literal_str(t);

	while ((ch = *p++) != '\0') {
		if (ch == '\\')
			fputc('\\', f);
		fputc(ch, f);
	}
}

void LLLdump_c_parse(FILE * f, TTterm id)
{
	fprintf(f, "TTparse");
	LLLdump_c_ident(f, id);
}

void LLLdump_c_typecheck(FILE * f, TTterm id)
{
	fprintf(f, "TTis");
/*	dump_c_type_body(f, id, spec);*/
	LLLdump_c_ident(f, id);
}

void LLLdump_c_build_name(FILE * f, TTterm id)
{
	fprintf(f, "TTbuild");
	LLLdump_c_ident(f, id);
}

void LLLdump_c_match_name(FILE * f, TTterm id)
{
	fprintf(f, "TTmatch");
	LLLdump_c_ident(f, id);
}

void LLLdump_c_fun_sym(FILE * f, TTterm id)
{
	fprintf(f, "TTfun");
	LLLdump_c_ident(f, id);
}

static BOOL is_epsilon(TTterm t)
{
	TTterm x, y;

	if(TTmatchLLL_basis(t, &x)
	   && TTequal_string(x, "%epsilon")) {
	   	return TRUE;
	} else if(TTmatchLLL_term(t, &x, &y)
		  && TTmatch_nil(y)) {
		return is_epsilon(x);
	}
	return FALSE;
}

static BOOL not_is_epsilon(TTterm t)
{
	return !is_epsilon(t);
}

static TTterm simplify_t(TTterm t)
{
	TTterm basis, rep_opt, xs, ys, x, u;

	if (TTmatchLLL_term(t, &basis, &rep_opt)
	    && !TTmatch_nil(rep_opt)
	    && is_epsilon(basis)) {
	    	return TTbuildLLL_term(basis, TTbuild_nil());
	} else if (TTmatchLLL_conjunction(t, &xs)) {
		ys = TTfilter(not_is_epsilon, xs);
		if (xs != ys && !TTmatch_nil(ys)) {
			return TTbuildLLL_conjunction(ys);
		}
	} else if (TTmatchLLL_group(t, &u, &x, &u)) {
		assert(TTmatchLLL_disjunction(x, &xs));
		if (TTmatch_singleton(xs, &x)) {
			assert(TTmatchLLL_conjunction(x, &xs));
			if (TTmatch_singleton(xs, &x)) {
				assert(TTmatchLLL_term(x, &basis, &rep_opt));
				if (TTmatch_nil(rep_opt)) {
					assert(TTmatchLLL_basis(basis, &x));
					return x;
				}
			}
		}
	}
	return t;
}

static TTterm lift_groups(TTterm t)
{
	TTterm x, xs, ys, zs, rep_opt, u;

	if (TTmatch_singleton(t, &x)) {
		assert(TTmatchLLL_conjunction(x, &xs));
		if (TTmatch_singleton(xs, &x)) {
			assert(TTmatchLLL_term(x, &x, &rep_opt));
			if (TTmatch_nil(rep_opt)) {
				assert(TTmatchLLL_basis(x, &x));
				if (TTmatchLLL_group(x, &u, &x, &u)) {
					assert(TTmatchLLL_disjunction(x, &xs));
					return xs;
				}
			}
		}
		return t;
	}
	if (!TTmatch_cons(t, &x, &ys)) {
		TTdump(t);
	}
	assert(TTmatch_cons(t, &x, &ys));
	zs = lift_groups(ys);
	assert(TTmatchLLL_conjunction(x, &xs));
	if (TTmatch_singleton(xs, &x)) {
		assert(TTmatchLLL_term(x, &x, &rep_opt));
		if (TTmatch_nil(rep_opt)) {
			assert(TTmatchLLL_basis(x, &x));
			if (TTmatchLLL_group(x, &u, &x, &u)) {
				assert(TTmatchLLL_disjunction(x, &xs));
				return TTconcat(xs, zs);
			}
		}
	}
	if (zs != ys) {
		return TTbuild_cons(x, zs);
	}
	return t;
}

static TTterm simplify_disjunction_t(TTterm t)
{
	TTterm xs, ys;

	if (TTmatchLLL_disjunction(t, &xs)) {
		ys = lift_groups(xs);
		if (xs != ys) {
			return TTbuildLLL_disjunction(ys);
		}
	}
	return t;
}

static TTterm insertid(TTterm x, TTterm ys)
{
	TTterm y, ys2, strx, stry;
	const char * px;
	const char * py;

	if (TTmatch_nil(ys)) {
		return TTbuild_singleton(x);
	}
	assert(TTmatch_cons(ys, &y, &ys2));
	assert(TTmatchLLL_ident(x, &strx));
	assert(TTmatchLLL_ident(y, &stry));
	assert(TTmatch_string(strx, &px));
	assert(TTmatch_string(stry, &py));
	if (strcmp(px, py) < 0) {
		return TTbuild_cons(x, ys);
	}
	return TTbuild_cons(y, insertid(x, ys2));
}

TTterm LLLsort_idents(TTterm t)
{
	TTterm xs, ys, x;

	xs = t;
	ys = TTbuild_nil();

	while (TTmatch_cons(xs, &x, &xs)) {
		ys = insertid(x, ys);
	}
	return ys;
}

static TTterm insert(TTterm x, TTterm ys)
{
	TTterm y, ys2, idx, idy, strx, stry, u;
	const char * px;
	const char * py;

	if (TTmatch_nil(ys)) {
		return TTbuild_singleton(x);
	}
	assert(TTmatch_cons(ys, &y, &ys2));
	assert(TTmatchLLL_rule(y, &idy, &u, &u, &u));
	assert(TTmatchLLL_rule(x, &idx, &u, &u, &u));
	assert(TTmatchLLL_ident(idx, &strx));
	assert(TTmatchLLL_ident(idy, &stry));
	assert(TTmatch_string(strx, &px));
	assert(TTmatch_string(stry, &py));
	if (strcmp(px, py) < 0) {
		return TTbuild_cons(x, ys);
	}
	return TTbuild_cons(y, insert(x, ys2));
}

TTterm LLLsort(TTterm t)
{
	TTterm xs, ys, x;

	assert(TTmatchLLL_specification(t, &xs));

	ys = TTbuild_nil();

	while (TTmatch_cons(xs, &x, &xs)) {
		ys = insert(x, ys);
	}
	return TTbuildLLL_specification(ys);
}

TTterm LLLsimplify(TTterm t)
{
	TTterm old, new;

	new = t;
	do {
		old = new;
		new = TTtransform_nonstop(old, simplify_t);
		new = TTtransform_nonstop(new, simplify_disjunction_t);
	} while (old != new);

	return new;
}


