
#include <stdio.h>
#include <stdarg.h>
#include <assert.h>
#include <string.h>
#include <stdlib.h>

#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"
#include "ttstack.h"
#include "tterror.h"
#include "ttparse.h"

#define BUFSZ 1024
#define TOP_N 10

extern int snprintf(char *s,  size_t  n,  const  char  *format, ...);

static TTbuf st_buf;
static int st_cur;
static int st_max;

/* Utilities. */

static void inc_cur(int n)
{
	st_cur += n;
	if (st_cur > st_max) {
		st_max = st_cur;
	}
/*
	assert(st_cur >= 0);
	assert(st_cur <= st_buf->size);
*/
}

TTterm TTparse_buffer(TTterm (*p)(void), TTbuf buf)
{
	TTterm pt;

	st_buf = buf;
	st_cur = 0;
	st_max = 0;

	pt = p();

	if (pt == 0 || !(st_cur == st_buf->size)) {
		TTparse_error();
	}

	return pt;
}

void TTparse_error(void)
{
	static char msg[BUFSZ];
	int line, column;

	TTbuffer_get_location(st_buf, st_max, &line, &column);
	snprintf(msg, BUFSZ, "Parse error in %s on line %d, column %d.\n",
		st_buf->name, line, column);
	TTerror(msg);
}

const char * TTparse_get_string(void)
{
	return st_buf->data + st_cur;
}

/* Basic combinators. */
TTtype_list TTparse_optional(TTterm(*parser) (void))
{
	TTterm t = parser();

	if (t) {
		return TTbuild_singleton(t);
	}
	return TTbuild_nil();
}

TTtype_list TTparse_plus(TTterm(*parser) (void))
{
	TTterm t = parser();

	if (t) {
		return TTbuild_cons(t, TTparse_star(parser));
	}
	return (TTtype_list) 0;
}

TTtype_list TTparse_star(TTterm(*parser) (void))
{
	TTterm t = parser();

	if (t) {
		return TTbuild_cons(t, TTparse_star(parser));
	}
	return (TTtype_list) TTbuild_nil();
}

TTterm TTparse_disjunction(TTfun fun, size_t n, TTterm(**parsers) (void))
{
	TTterm t;
	size_t i;

	for (i = 0; i < n; i++) {
		if ((t = parsers[i] ()) != 0) {
			return TTbuild_term(fun, &t);
		}
	}
	return 0;
}

TTtype_list TTparse_alternation_plus(TTterm(*p1) (void), TTterm(*p2) (void))
{
	TTterm hd = p1();
	TTterm sep;
	TTtype_list tl;
	int save_cur;

	if (hd) {
 		save_cur = st_cur;
		sep = p2();
		if (sep) {
			tl = TTparse_alternation_plus(p1, p2);
			if (tl) {
				return TTbuild_cons(TTbuild_sandwich(TTbuild_empty_string(), hd, sep), tl);
			}
		}
		st_cur = save_cur;
		return TTbuild_singleton(hd);
	}
	return (TTtype_list) 0;
}

TTtype_list TTparse_alternation_star(TTterm(*p1) (void), TTterm(*p2) (void))
{
	TTterm result;

	result = TTparse_alternation_plus(p1, p2);
	if (!result) {
		result = TTbuild_nil();
	}
	return result;
}

TTterm TTparse_conjunction(TTfun fun, TTterm(**parsers) (void))
{
	size_t i;
	TTterm *args = (TTterm *) parsers;
	int save_cur = st_cur;

	for (i = 0; i < fun->arity; i++) {
		if(!(args[i] = parsers[i] ())) {
			st_cur = save_cur;
			return 0;
		}
	}
	return TTbuild_term(fun, args);
}

TTtype_string TTparse_word(size_t len)
{
	TTterm t = TTbuild_string_n(st_buf->data + st_cur, len);
	inc_cur((int)len);
	return t;
}

/* Advanced combinators. */
BOOL TTparse_guard(TTterm(*parser) (void))
{
	TTterm t;
	int save_cur = st_cur;

	t = parser();
	st_cur = save_cur;
	return t != 0;
}

