
#include <stdio.h>
#include <malloc.h>
#include <assert.h>
#include "ttterm.h"
#include "ttstrat.h"

/* Shortcuts */
static int init = 0;

static TTterm st_succeed;
static TTterm st_fail;

static TTterm args[10];

/* The basic strategies as specified on page 5 of ESP. */
static TTterm apply_adhoc(const TTterm * s, TTterm t, const TTstrat_env env)
{
	int n;

	assert(TTmatch_int(s[0], &n));
	return ((TTreduce_f) n) (t);
}

static TTterm apply_succeed(const TTterm * s, TTterm t, const TTstrat_env env)
{
	if (env) { return env->empty; }
	return t;
}

static TTterm apply_fail(const TTterm * s, TTterm t, const TTstrat_env env)
{
	return 0;
}

static TTterm apply_both_tp(const TTterm * s, TTterm t)
{
	TTterm t1;

	t1 = TTstrat_apply_tp(s[0], t);
	if (!t1) return 0;
	return TTstrat_apply_tp(s[1], t1);
}

static TTterm apply_both_tu(const TTterm * s, TTterm t, const TTstrat_env env)
{
	TTterm t1, t2;

	t1 = TTstrat_apply(s[0], t, env);
	if (!t1) return 0;
	t2 = TTstrat_apply(s[1], t, env);
	if (!t2) return 0;
	return env->comb(t1, t2);
}

static TTterm apply_both(const TTterm * s, TTterm t, const TTstrat_env env)
{
	return env ? apply_both_tu(s, t, env) : apply_both_tp(s, t);
}

static TTterm apply_choice(const TTterm * s, TTterm t, const TTstrat_env env)
{
	TTterm tmp = TTstrat_apply(s[0], t, env);
	if (tmp) { return tmp; }
	return TTstrat_apply(s[1], t, env);
}

static TTterm apply_all_tp(const TTterm * s, TTterm t)
{
	TTterm small_args[7];
        TTterm result;
	TTfun fun;
	int arity;
        int i;

        if (!TTis_node(t)) return t;

	fun = t->fun;
	arity = fun->arity;

	if (arity <= 7) {
		for (i = 0; i < arity; i++) {
			small_args[i] = TTstrat_apply_tp(s[0], t->u.args[i]);
			if (!small_args[i]) {
				return 0;
			}
		}
		for (i = 0; i < arity; i++) {
			if (small_args[i] != t->u.args[i]) {
				return TTbuild_term(fun, small_args);
			}
		}
		return t;
	} else {
	        result = TTalloc(fun);
		for (i = 0; i < arity; i++) {
			result->u.args[i] = TTstrat_apply_tp(s[0], t->u.args[i]);
			if (!result->u.args[i]) {
				TTfree(result);
				return 0;
			}
		}
		for (i = 0; i < arity; i++) {
			if (result->u.args[i] != t->u.args[i]) {
				return TTfinish(result);
			}
		}
		TTfree(result);
		return t;
	}
}

static TTterm apply_all_tu(const TTterm * s, TTterm t, const TTstrat_env env)
{
        TTterm result, t2;
        int i;

        if (!TTis_node(t)) return env->empty;

        result = env->empty;
        for (i = 0; i < t->fun->arity; i++) {
                t2 = TTstrat_apply(s[0], t->u.args[i], env);
                if (!t2) return 0;
                result = env->comb(result, t2);
        }
        return result;
}

static TTterm apply_all(const TTterm * s, TTterm t, const TTstrat_env env)
{
	return env ? apply_all_tu(s, t, env) : apply_all_tp(s, t);
}

static TTterm apply_one_tp(const TTterm * s, TTterm t)
{
	TTterm small_args[7];
        TTterm result, t2;
	TTfun fun;
	int arity;
        int i, j;

        if (!TTis_node(t)) return 0;

	fun = t->fun;
	arity = fun->arity;

        for (i = 0; i < arity; i++) {
                t2 = TTstrat_apply_tp(s[0], t->u.args[i]);
                if (t2) {
                        result = TTalloc(fun);
                        for (j = 0; j < arity; j++) {
                                result->u.args[j] = t->u.args[j];
                        }
                        result->u.args[i] = t2;
                        TTfinish(result);
                        return result;
                }
        }
        return 0;
}

static TTterm apply_one_tu(const TTterm * s, TTterm t, const TTstrat_env env)
{
        TTterm t2;
        int i;

        if (!TTis_node(t)) return 0;

        for (i = 0; i < t->fun->arity; i++) {
                t2 = TTstrat_apply(s[0], t->u.args[i], env);
                if (t2) return t2;
        }
        return 0;
}

static TTterm apply_one(const TTterm * s, TTterm t, const TTstrat_env env)
{
	return env ? apply_one_tu(s, t, env) : apply_one_tp(s, t);
}

/* API, making and applying strategies. */
TTterm TTbuild_adhoc(TTreduce_f r)
{
	args[0] = TTbuild_int((int) r);
	return TTbuild_term(TTfun_adhoc, args);
}

TTterm TTbuild_succeed(void)
{
	return st_succeed;
}

TTterm TTbuild_fail(void)
{
	return st_fail;
}

TTterm TTbuild_both(TTterm s1, TTterm s2)
{
	args[0] = s1;
	args[1] = s2;
	return TTbuild_term(TTfun_both, args);
}

TTterm TTbuild_choice(TTterm s1, TTterm s2)
{
	args[0] = s1;
	args[1] = s2;
	return TTbuild_term(TTfun_choice, args);
}

TTterm TTbuild_all(TTterm s)
{
	args[0] = s;
	return TTbuild_term(TTfun_all, args);
}

TTterm TTbuild_one(TTterm s)
{
	args[0] = s;
	return TTbuild_term(TTfun_one, args);
}

TTfun TTstrat_fun(const char * name, size_t arity, TTapply_strat_f f)
{
	TTfun fun;

	fun = TTbuild_fun(name, arity);
	fun->reduce = (TTreduce_f)f;
	return fun;
}

void TTstrat_init(void)
{
	if (init == 0) {
		TTterm_init();
		TTfun_adhoc = TTstrat_fun("adhoc", 1, apply_adhoc);
		TTfun_succeed = TTstrat_fun("succeed", 0, apply_succeed);
		TTfun_fail = TTstrat_fun("fail", 0, apply_fail);
		TTfun_both = TTstrat_fun("both", 2, apply_both);
		TTfun_choice = TTstrat_fun("choice", 2, apply_choice);
		TTfun_all = TTstrat_fun("all", 1, apply_all);
		TTfun_one = TTstrat_fun("one", 1, apply_one);

		st_succeed = TTbuild_term(TTfun_succeed, args);
		st_fail = TTbuild_term(TTfun_fail, args);
	}
	init++;
}

TTterm TTstrat_apply(TTterm s, TTterm t, TTstrat_env env)
{
	TTapply_strat_f f;

	if (!t) { return t; }
	f = (TTapply_strat_f)s->fun->reduce;
	return f((const TTterm *) (s->u.args), t, env);
}

TTterm TTstrat_apply_tp(TTterm s, TTterm t)
{
	TTapply_strat_f f;

	if (!t) { return t; }
	f = (TTapply_strat_f)s->fun->reduce;
	return f((const TTterm *) (s->u.args), t, 0);
}

