#include <errno.h>
#include <stdlib.h>
#include <assert.h>
#include <stdio.h>
#include <string.h>
#include <sys/types.h>
#include "ttterm.h"
#include "gc.h"

enum enum_content_type {
	TT_SUBTERMS = 1,
	TT_INT = 2,
	TT_STR = 3
};

#define HEADER_SIZE (sizeof(struct TTterm_struct) - 4)
#define MAX_FUNS_TABLE_SIZE 10001

static int init = 0;
static TTfun st_funs_table[MAX_FUNS_TABLE_SIZE];

/* Don't need a good hash function yet. */
static int tt_hash(const unsigned char * str, size_t length)
{
	int i, result;

	result = 0;
	for (i = 0; i < length; i++) {
		result += str[i];
	}
	return result;
}

BOOL TTis_node(TTterm t)
{
	return (t->fun->type == TT_SUBTERMS);
}

TTterm TTbuild_term(TTfun fun, const TTterm * args)
{
	size_t memsize = fun->arity * 4;
	TTterm t = (TTterm) GC_malloc(HEADER_SIZE + memsize);
	t->fun = fun;
	memcpy(t->u.args, args, memsize);
	return t;
}

TTterm TTalloc(TTfun fun)
{
	size_t memsize = fun->arity * 4;
	TTterm t = (TTterm) GC_malloc(HEADER_SIZE + memsize);
	t->fun = fun;
	return t;
}

void TTfree(TTterm t)
{
	GC_free(t);
}

TTterm TTfinish(TTterm t)
{
	return t;
}

static TTfun build_fun(size_t arity, int typ, const char *name)
{
	TTfun result, x;
	int first;

	first = tt_hash((const unsigned char *)name, strlen(name)) % MAX_FUNS_TABLE_SIZE;
	x = st_funs_table[first];

	/* Lookup the fun in the st_funs_table. */
	while (x != 0) {
		if (strcmp(x->name, name) == 0) {
			/* If the fun already exists return it. */
			assert(x->arity == arity);
			assert(x->type == typ);
			return x;
		} else {
			/* Keep looking. */
			x = x->next;
		}
	}
	/* The fun did not already exist, so make it. */
	result = (TTfun) GC_malloc(sizeof(struct TTfun_struct));
	result->arity = arity;
	result->type = typ;
	result->name = (char *) GC_malloc(strlen(name) + 1);
	strcpy(result->name, name);

	/* Put it in the hashtable. */
	result->next = st_funs_table[first];
	st_funs_table[first] = result;

	return result;
}

TTfun TTbuild_fun(const char *name, size_t arity)
{
	return build_fun(arity, TT_SUBTERMS, name);
}

BOOL TTequal(TTterm t1, TTterm t2)
{
	size_t i;
	TTterm *args1 = t1->u.args;
	TTterm *args2 = t2->u.args;
	TTfun fun;

	if (t1 == t2)
		return TRUE;
	fun = t1->fun;
	if (fun != t2->fun)
		return FALSE;

	if (TTis_node(t1)) {
		for (i = 0; i < fun->arity; i++) {
			if (!TTequal(*args1++, *args2++))
				return FALSE;
		}
		return TRUE;
	} else if (fun == TTfun_string) {
		return strcmp(t1->u.str, t2->u.str) == 0;
	}

	assert(fun == TTfun_int);
	return t1->u.val == t2->u.val;
}

BOOL TTequal_fun(TTfun f1, TTfun f2)
{
	return f1 == f2;
}

TTterm TTbuild_int(int n)
{
	TTterm result;
	result = TTalloc(TTfun_int);
	result->u.val = n;
	return TTfinish(result);
}

BOOL TTmatch_int(TTterm t, int *x)
{
	if (t->fun == TTfun_int) {
		*x = t->u.val;
		return TRUE;
	}
	return FALSE;
}

TTtype_string TTbuild_string(const char * str)
{
	size_t memsize = strlen(str) + 1;
	TTterm t = (TTterm) GC_malloc(HEADER_SIZE + memsize);
	t->fun = TTfun_string;
	strcpy(t->u.str, str);
	return (TTtype_string)t;
}

BOOL TTmatch_string(TTtype_string t, const char ** x)
{
	if (((TTterm) t)->fun == TTfun_string) {
		*x = t->u.str;
		return TRUE;
	}
	return FALSE;
}

TTterm TTalloc_string(int len)
{
	size_t memsize = len + 1;
	TTterm t = (TTterm) GC_malloc(HEADER_SIZE + memsize);
	t->fun = TTfun_string;
	return t;
}

void TTfree_string(TTterm t)
{
	GC_free(t);
}

TTterm TTfinish_string(TTterm t)
{
	return t;
}

void TTdump(TTterm t)
{
	size_t i;
	int n;
	const TTterm *args;
	TTfun fun = t->fun;
	const char * str;

	fprintf(stderr, "----------------------------------------\n");
	fprintf(stderr, "dump term: %d\n", (int) t);
	fprintf(stderr, "fun:     %s\n", fun->name);

	if (TTmatch_int(t, &n)) {
		fprintf(stderr, "int:     %d\n", n);
	} else if (TTmatch_string((TTtype_string)t, &str)) {
		fprintf(stderr, "str:     %s\n", str);
	} else {
		assert(TTis_node(t));
		args = t->u.args;
		for (i = 0; i < fun->arity; i++) {
			fprintf(stderr, "sub[%u]:  %s\n",
				(unsigned int) i, args[i]->fun->name);
		}
	}
}

void TTterm_init(void)
{
	int i;

	if (init == 0) {
		/* Predefined funs. */
		TTfun_int = build_fun(1, TT_INT, "int");
		TTfun_string = build_fun(0, TT_STR, "string");

		for (i = 0; i < MAX_FUNS_TABLE_SIZE; i++) {
			st_funs_table[i] = 0;
		}
	}
	init++;
}

