
#include <ctype.h>
#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <sys/types.h>
#include "ttterm.h"
#include "ttbuffer.h"
#include "ttutils.h"

static TTtype_string st_empty_string;
static TTtype_list st_nil;
static TTbuf st_buf;
static int init = 0;
static TTterm args[10];

void TTutils_init(void)
{
	if (init == 0) {
		TTterm_init();
		TTfun_nil = TTbuild_fun("nil", 0);
		TTfun_cons = TTbuild_fun("cons", 2);
		TTfun_tuple = TTbuild_fun("tuple", 2);
		TTfun_sandwich = TTbuild_fun("sandwich", 3);
/*		TTfun_anno = TTbuild_fun("anno", 2);*/
		TTfun_layout = TTbuild_fun("layout", 2);
		st_nil = (TTtype_list) TTbuild_term(TTfun_nil, 0);
		st_empty_string = (TTtype_string) TTbuild_string("");
	}
	init++;
}

/* Tuples */
TTtype_tuple TTbuild_tuple(TTterm t1, TTterm t2)
{
	args[0] = t1;
	args[1] = t2;
	return (TTtype_tuple) TTbuild_term(TTfun_tuple, args);
}

BOOL TTmatch_tuple(TTtype_tuple t, TTterm * t1, TTterm * t2)
{
	TTterm tt = (TTterm) t;
	if (TTequal_fun(tt->fun, TTfun_tuple)) {
		*t1 = tt->u.args[0];
		*t2 = tt->u.args[1];
		return TRUE;
	}
	return FALSE;
}

BOOL TTis_tuple(TTterm t)
{
	return TTequal_fun(t->fun, TTfun_tuple);
}

/* Lists */
TTtype_list TTconcat(TTtype_list x, TTtype_list y)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_cons(x, &hd, &tl)) {
		return TTbuild_cons(hd, TTconcat(tl, y));
	}
	assert(TTmatch_nil(x));
	return y;
}

int TTlength(TTtype_list t)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_cons(t, &hd, &tl)) {
		return 1 + TTlength(tl);
	}
	assert(TTmatch_nil(t));
	return 0;
}

TTtype_list TTbuild_cons(TTterm hd, TTtype_list tl)
{
	args[0] = hd;
	args[1] = tl;
	return TTbuild_term(TTfun_cons, args);
}

TTtype_list TTbuild_nil(void)
{
	return st_nil;
}

BOOL TTis_list(TTterm t)
{
	return TTequal_fun(t->fun, TTfun_cons) || TTequal_fun(t->fun, TTfun_nil);
}

BOOL TTmatch_cons(TTtype_list t, TTterm * hd, TTtype_list * tl)
{
	TTterm tt = (TTterm) t;
	if (TTequal_fun(tt->fun, TTfun_cons)) {
		*hd = tt->u.args[0];
		*tl = (TTtype_list) tt->u.args[1];
		return TRUE;
	}
	return FALSE;
}

BOOL TTmatch_nil(TTtype_list t)
{
	return TTequal_fun(((TTterm) t)->fun, TTfun_nil);
}

TTterm TThead(TTtype_list t)
{
	TTterm hd;
	TTtype_list tl;

	assert(TTmatch_cons(t, &hd, &tl));
	return hd;
}

TTtype_list TTtail(TTtype_list t)
{
	TTterm hd;
	TTtype_list tl;

	assert(TTmatch_cons(t, &hd, &tl));
	return tl;
}

TTtype_list TTrepeat(int n, TTterm x)
{
	TTtype_list result;
	int i;

	result = st_nil;
	for (i = 0; i < n; i++) {
		result = TTbuild_cons(x, result);
	}
	return result;
}

TTtype_list TTlinit(TTtype_list xs)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_cons(xs, &hd, &tl)) {
		if (TTmatch_nil(tl)) {
			return st_nil;
		} else {
			return TTbuild_cons(hd, TTlinit(tl));
		}
	}
	assert(FALSE);
	return (TTtype_list) 0;
}

TTtype_list TTbuild_singleton(TTterm x)
{
	return TTbuild_cons(x, st_nil);
}

BOOL TTmatch_singleton(TTtype_list x, TTterm * y)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_cons(x, &hd, &tl)) {
		if (TTmatch_nil(tl)) {
			*y = hd;
			return TRUE;
		}
	}
	return FALSE;
}

TTtype_list TTreverse(TTtype_list xs)
{
	TTterm hd;
	TTtype_list result;

	result = TTbuild_nil();
	while (TTmatch_cons(xs, &hd, &xs)) {
		result = TTbuild_cons(hd, result);
	}
	return result;
}

TTtype_list TTtake(int n, TTtype_list xs)
{
	TTterm hd, tl;

	if (n <= 0 || TTmatch_nil(xs)) {
		return TTbuild_nil();
	}
	assert(TTmatch_cons(xs, &hd, &tl));
	return TTbuild_cons(hd, TTtake(n-1, tl));
}

TTtype_list TTdrop(int n, TTtype_list xs)
{
	TTterm u;

	while (n > 0 && TTmatch_cons(xs, &u, &xs)) {
		n--;
	}
	return xs;
}

/* Sets */
BOOL TTelem(TTterm x, TTtype_list xs)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_nil(xs)) {
		return FALSE;
	} else {
		assert(TTmatch_cons(xs, &hd, &tl));
		if (TTequal(x, hd))
			return TRUE;
		return TTelem(x, tl);
	}
}

BOOL TTsubset(TTtype_list xs, TTtype_list ys)
{
	TTterm hd;

	while (TTmatch_cons(xs, &hd, &xs)) {
		if (!TTelem(hd, ys)) {
			return FALSE;
		}
	}
	assert(TTmatch_nil(xs));
	return TRUE;
}

TTtype_list TTunion(TTtype_list xs, TTtype_list ys)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_nil(xs)) {
		return ys;
	} else {
		assert(TTmatch_cons(xs, &hd, &tl));
		if (TTelem(hd, ys)) {
			return TTunion(tl, ys);
		} else {
			return TTbuild_cons(hd, TTunion(tl, ys));
		}
	}
}

TTtype_list TTdiff(TTtype_list xs, TTtype_list ys)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_nil(xs)) {
		return st_nil;
	} else {
		assert(TTmatch_cons(xs, &hd, &tl));
		if (TTelem(hd, ys)) {
			return TTdiff(tl, ys);
		} else {
			return TTbuild_cons(hd, TTdiff(tl, ys));
		}
	}
}

TTtype_list TTmap(TTmap_f f, TTtype_list xs)
{
	TTterm hd, hd2;
	TTtype_list tl, tl2;

	if (TTmatch_nil(xs)) {
		return st_nil;
	} else {
		assert(TTmatch_cons(xs, &hd, &tl));
		hd2 = f(hd);
		tl2 = TTmap(f, tl);
		if (hd2 == hd && tl2 == tl)
			return xs;
		return TTbuild_cons(hd2, tl2);
	}
}

void TTmap_void(TTmapvoid_f f, TTtype_list xs)
{
	TTterm hd;
	TTtype_list tl;

	if (TTmatch_cons(xs, &hd, &tl)) {
		f(hd);
		TTmap_void(f, tl);
	} else {
		assert(TTmatch_nil(xs));
	}
}

TTtype_list TTfilter(TTfilter_f f, TTtype_list xs)
{
	TTterm hd;
	TTtype_list tl, tl2;
	BOOL keep;

	if (TTmatch_nil(xs)) {
		return xs;
	} else {
		assert(TTmatch_cons(xs, &hd, &tl));
		keep = f(hd);
		tl2 = TTfilter(f, tl);
		if (keep) {
			if (tl2 == tl) {
				return xs;
			}
			return TTbuild_cons(hd, tl2);
		}
		return tl2;
	}
}

TTtype_list TTfilter_map(TTmap_f f, TTtype_list t)
{
	TTterm hd, tl;

	if (TTmatch_cons(t, &hd, &tl)) {
		hd = f(hd);
		tl = TTfilter_map(f, tl);
		if (!hd)
			return tl;
		return TTbuild_cons(hd,tl);
	}
	assert(TTmatch_nil(t));
	return t;
}

/* FiniteMaps */
TTtype_fm TTempty_fm(void)
{
	return (TTtype_fm) st_nil;
}

TTtype_fm TTadd_to_fm(TTtype_fm fm, TTterm key, TTterm elt)
{
	return (TTtype_fm) TTbuild_cons((TTterm) TTbuild_tuple(key, elt),
				     (TTtype_list) fm);
}

TTterm TTlookup_fm(TTtype_fm fm, TTterm key)
{
	TTterm hd;
	TTtype_list tl;
	TTterm key2, elt2;

	if (TTmatch_nil((TTtype_list) fm)) {
		return 0;
	}
	assert(TTmatch_cons((TTtype_list) fm, &hd, &tl));
	assert(TTmatch_tuple((TTtype_tuple) hd, &key2, &elt2));
	if (TTequal(key, key2))
		return elt2;
	return TTlookup_fm((TTtype_fm) tl, key);
}

/* Strings */
size_t TTstrlen(TTtype_string t)
{
	const char * str;

	assert(TTmatch_string(t, &str));
	return strlen(str);
}

void TTdump_string(FILE * f, TTtype_string t)
{
	const char * str;

	assert(TTmatch_string(t, &str));
	fputs(str, f);
}

TTtype_string TTunquote(TTtype_string t)
{
	int len;
	const char * str;
	TTterm result;

	assert(TTmatch_string(t, &str));
	len = strlen(str);
	assert(len >= 2);
	result = TTalloc_string(len - 2);
	strncpy(result->u.str, str + 1, (size_t) len - 2);
	result->u.str[len - 2] = '\0';
	return TTfinish_string(result);
}

TTtype_string TTquote(int open, int close, TTtype_string t)
{
	int len;
	const char * str;
	TTterm result;

	assert(TTmatch_string(t, &str));
	len = strlen(str);
	result = TTalloc_string(len + 2);
	strncpy(result->u.str + 1, str, (size_t)len);
	result->u.str[0] = open;
	result->u.str[len + 1] = close;
	result->u.str[len + 2] = '\0';
	return TTfinish_string(result);
}

TTtype_string TTto_upper(TTtype_string t)
{
	int i;
	int len = TTstrlen(t);
	TTterm result = TTalloc_string(len);

	for (i = 0; i < len; i++) {
		result->u.str[i] = toupper(t->u.str[i]);
	}
	return TTfinish_string(result);
}

TTtype_string TTcapitalize(TTtype_string t)
{
	int len = TTstrlen(t);
	TTterm result = TTalloc_string(len);

	strcpy(result->u.str, t->u.str);
	result->u.str[0] = toupper(t->u.str[0]);
	return TTfinish_string(result);
}

TTtype_string TTreplace(int c1, int c2, TTtype_string t)
{
	int i, ch;
	int len = TTstrlen(t);
	TTterm result = TTalloc_string(len);

	for (i = 0; i < len; i++) {
		ch = t->u.str[i];
		result->u.str[i] = ch == c1 ? c2 : ch;
	}
	return TTfinish_string(result);
}

BOOL TTequal_string(TTtype_string t, const char *name)
{
	const char * str;

	if (TTmatch_string(t, &str)) {
		return strcmp(str, name) == 0;
	}
	return FALSE;
}

BOOL TTequal_string_ci(TTtype_string t, const char *name)
{
	const char * p;
	const char * q;

	assert(TTmatch_string(t, &p));
	q = name;
	while (1) {
		if (tolower(*p) != tolower(*q)) {
			return FALSE;
		}
		if (*p == '\0' && *q == '\0') {
			return TRUE;
		}
		p++;
		q++;
	}
}

TTtype_string TTbuild_empty_string(void)
{
	return st_empty_string;
}

TTtype_string
TTconcat_string(TTtype_string x, TTtype_string y)
{
	int len_x, len_y;
	const char * str_x;
	const char * str_y;
	TTterm result;

	assert(TTmatch_string(x, &str_x));
	assert(TTmatch_string(y, &str_y));
	len_x = strlen(str_x);
	len_y = strlen(str_y);
	result = TTalloc_string(len_x + len_y);
	strncpy(result->u.str, str_x, (size_t)len_x);
	strncpy(result->u.str + len_x, str_y, (size_t)len_y);
	result->u.str[len_x + len_y] = '\0';
	return TTfinish_string(result);
}

TTtype_string
TTbuild_string_n(const char * str, size_t len)
{
	TTterm result = TTalloc_string((int)len);
	strncpy(result->u.str, str, len);
	result->u.str[len] = '\0';
	return TTfinish_string(result);
}

/* Traversals */

/* Fast term equality check for transform and transform_nonstop. */
static BOOL equal(int arity, TTterm * args1, TTterm * args2)
{
	return memcmp(args1, args2, arity * sizeof(TTterm)) == 0;
}

TTterm TTtransform(TTterm t, TTreduce_f f)
{
	/* 7 is an arbitrary choice, anything higher then 2 is fine. */
	TTterm small_args[7];
	TTterm result;
	TTfun fun;
	int arity;
	size_t i;

	result = f(t);
	if (result) {
		return result;
	} else if (TTis_node(t)) {
		fun = t->fun;
		arity = fun->arity;
		if (arity <= 7) {
			for (i = 0; i < arity; i++) {
				small_args[i] = TTtransform(t->u.args[i], f);
			}
			if (equal(arity, small_args, t->u.args)) {
				return t;
			}
			return TTbuild_term(fun, small_args);
		} else {
			result = TTalloc(fun);
			for (i = 0; i < arity; i++) {
				result->u.args[i] = TTtransform(t->u.args[i], f);
			}
			if (equal(arity, result->u.args, t->u.args)) {
				TTfree(result);
				return t;
			}
			return TTfinish(result);
		}
	}
	return t;
}

TTterm TTtransform_nonstop(TTterm t, TTreduce_f f)
{
	TTterm small_args[7];
	TTterm result;
	TTfun fun;
	int arity;
	size_t i;

	t = f(t);
	if (TTis_node(t)) {
		fun = t->fun;
		arity = fun->arity;
		if (arity <= 7) {
			for (i = 0; i < arity; i++) {
				small_args[i] = TTtransform_nonstop(t->u.args[i], f);
			}
			if (equal(arity, small_args, t->u.args)) {
				return t;
			}
			return TTbuild_term(fun, small_args);
		} else {
			result = TTalloc(fun);
			for (i = 0; i < arity; i++) {
				result->u.args[i] = TTtransform_nonstop(t->u.args[i], f);
			}
			if (equal(arity, result->u.args, t->u.args)) {
				TTfree(result);
				return t;
			}
			return TTfinish(result);
		}
	}
	return t;
}

TTterm TTanalyze(TTterm t,
		 TTtrafo_f f,
		 TTterm (*comb)(TTterm, TTterm),
		 TTterm empty)
{
	size_t i;
	TTterm result, tmp;

	result = f(t);
	if (result) {
		return result;
	} else if (TTis_node(t)) {
		result = empty;
		for (i = 0; i < t->fun->arity; i++) {
			tmp = TTanalyze(t->u.args[i], f, comb, empty);
			result = comb(result, tmp);
		}
		return result;
	}
	return empty;
}

void TTtraverse(TTterm t, TTtrav_f f)
{
	size_t i;
	TTterm * p;

	if (!f(t) && TTis_node(t)) {
		p = t->u.args;
		for (i = 0; i < t->fun->arity; i++) {
			TTtraverse(*p++, f);
		}
	}
}

static void traverse_line(TTterm t, TTtrav_line_f f, int * line, int * col)
{
	size_t i;
	const char *p;
	int ch;

	f(t, *line, *col);

	if (TTmatch_string(t, &p)) {
		while ((ch = *p++) != '\0') {
			if (ch == '\n') {
				(*line)++;
				*col = 1;
			} else {
				(*col)++;
			}
		}
	} else if (TTis_node(t)) {
		for (i = 0; i < t->fun->arity; i++) {
			traverse_line(t->u.args[i], f, line, col);
		}
	}
}

void TTtraverse_line(TTterm t, TTtrav_line_f f, int line, int col)
{
	int l = line;
	int c = col;

	traverse_line(t, f, &l, &c);
}

TTterm TTselect(TTterm t, TTtrafo_f f)
{
	TTterm result;
	size_t i;
	TTterm tmp;

	result = f(t);
	if (result) {
		return result;
	} else if (TTis_node(t)) {
		for (i = 0; i < t->fun->arity; i++) {
			tmp = TTselect(t->u.args[i], f);
			if (tmp) {
				return tmp;
			}
		}
	}
	return 0;
}

/* Text IO */
/*
static BOOL dump_string_t(TTterm t)
{
	if (t->fun == TTfun_string) {
		TTdump_string(st_file, (TTtype_string) t);
		fprintf(st_file, "\n");
		return TRUE;
	}
	return FALSE;
}

void TTdump_strings(FILE * f, TTterm t)
{
	st_file = f;
	TTtraverse(t, dump_string_t);
}
*/

static BOOL unparse(TTterm t)
{
	const char * p;

	if (TTmatch_string(t, &p)) {
		TTbuffer_append(st_buf, p);
		return TRUE;
	}
	return FALSE;
}

void TTunparse(TTbuf buf, TTterm t)
{
	st_buf = buf;
	TTtraverse(t, unparse);
}

void TTshow(FILE *f, TTterm t)
{
	const char * p;
	int i, n;
	TTterm hd, tl;
	BOOL first;

	if (TTmatch_string(t, &p)) {
		fprintf(f, "\"%s\"", p);
	} else if (TTmatch_int(t, &n)) {
		fprintf(f, "%d", n);
	} else if (TTmatch_cons(t, &hd, &tl)) {
		fprintf(f, "[");
		TTshow(f, hd);
		while (TTmatch_cons(tl, &hd, &tl)) {
			fprintf(f, ",");
			TTshow(f, hd);
		}
		fprintf(f, "]");
	} else if (TTmatch_nil(t)) {
		fprintf(f, "[]");
	} else {
		fprintf(f, "%s(", t->fun->name);
		first = TRUE;
		for (i = 0; i < t->fun->arity; i++) {
			if (!first) {
				fprintf(f, ",");
			}
			first = FALSE;
			TTshow(f, t->u.args[i]);
		}
		fprintf(f, ")");
	}
}

/* Special functions to deal with layout in syntax trees. */
TTtype_sandwich TTbuild_sandwich(TTterm t1, TTterm t2, TTterm t3)
{
	args[0] = t1;
	args[1] = t2;
	args[2] = t3;
	return (TTtype_sandwich) TTbuild_term(TTfun_sandwich, args);
}

BOOL TTmatch_sandwich(TTtype_sandwich t, TTterm * t1, TTterm * t2, TTterm * t3)
{
	TTterm tt = (TTterm) t;
	if (TTequal_fun(tt->fun, TTfun_sandwich)) {
		*t1 = tt->u.args[0];
		*t2 = tt->u.args[1];
		*t3 = tt->u.args[2];
		return TRUE;
	}
	return FALSE;
}

TTterm TTunsandwich(TTterm t)
{
	TTterm u, result;

	result = t;
	while (TTmatch_sandwich(result, &u, &result, &u)) {
	}
	return result;
}

TTterm TTunsandwich_rec(TTterm t)
{
	return TTtransform_nonstop(t, TTunsandwich);
}

BOOL TTutils_equal(TTterm t1, TTterm t2)
{
	size_t i;

	t1 = TTunsandwich(t1);
	t2 = TTunsandwich(t2);

	if (t1->fun != t2->fun) return FALSE;

	if (TTis_node(t1)) {
		for (i = 0; i < t1->fun->arity; i++) {
			if (!TTutils_equal(t1->u.args[i], t2->u.args[i])) {
				return FALSE;
			}
		}
		return TRUE;
	}
	return TTequal(t1, t2);
}

/* Annotations. */
/*
TTtype_anno TTbuild_anno(TTterm t1, TTterm t2)
{
	args[0] = t1;
	args[1] = t2;
	return (TTtype_anno) TTbuild_term(TTfun_anno, args);
}

BOOL TTmatch_anno(TTtype_anno t, TTterm * t1, TTterm * t2)
{
	TTterm tt = (TTterm) t;
	if (TTequal_fun(tt->fun, TTfun_anno)) {
		*t1 = tt->u.args[0];
		*t2 = tt->u.args[1];
		return TRUE;
	}
	return FALSE;
}

TTterm TTunanno(TTterm t)
{
	TTterm u, result;

	result = t;
	while (TTmatch_anno(result, &u, &result)) {
	}
	return result;
}

TTterm TTunanno_rec(TTterm t)
{
	return TTtransform_nonstop(t, TTunanno);
}

BOOL TTequal_ast(TTterm t1, TTterm t2)
{
	size_t i;

	t1 = TTunanno(t1);
	t2 = TTunanno(t2);

	if (t1->fun != t2->fun) return FALSE;

	if (TTis_node(t1)) {
		for (i = 0; i < t1->fun->arity; i++) {
			if (!TTequal(t1->u.args[i], t2->u.args[i])) {
				return FALSE;
			}
		}
		return TRUE;
	}
	return TTequal(t1, t2);
}
*/

/* Shallow parameterized type checking. */
BOOL TTtypecheck_optional(TTterm t, BOOL (*f)(TTterm))
{
	TTterm x;

	if (TTmatch_nil(t)) return TRUE;
	if (TTmatch_singleton(t, &x)) {
		return f(x);
	}
	return FALSE;
}

BOOL TTtypecheck_plus(TTterm t, BOOL (*f)(TTterm))
{
	TTterm hd, tl;

	if (TTmatch_nil(t)) return FALSE;
	if (TTmatch_cons(t, &hd, &tl)) {
		return f(hd);
	}
	return FALSE;
}

BOOL TTtypecheck_star(TTterm t, BOOL (*f)(TTterm))
{
	TTterm hd, tl;

	if (TTmatch_nil(t)) return TRUE;
	if (TTmatch_cons(t, &hd, &tl)) {
		return f(hd);
	}
	return FALSE;
}

BOOL TTtypecheck_keyword(TTterm t, const char * str)
{
	if (t->fun != TTfun_string) return FALSE;
	return TTequal_string_ci(t, str);
}

BOOL TTtypecheck_literal(TTterm t, const char * str)
{
	if (t->fun != TTfun_string) return FALSE;
	return TTequal_string_ci(t, str);
}

/* The weight of the term: 1 + the weight of its children. */
int TTweight(TTterm t)
{
	int i;
	int count = 1;

	if (TTis_node(t)) {
		for (i = 0; i < t->fun->arity; i++) {
			count += TTweight(t->u.args[i]);
		}
	}
	return count;
}

